# -*- encoding: utf-8 -*-
# Copyright (c) 2019 European Organization for Nuclear Research (CERN)
#
# Authors: Corne Lukken <info@dantalion.nl>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from watcher._i18n import _
from watcher.common import exception
from watcher.decision_engine.datasources import base


class BaseGrafanaTranslator(object):
    """Grafana translator baseclass to use with grafana for different databases

    Specific databasses that are proxied through grafana require some
    alterations depending on the database.
    """

    """
    data {
        metric: name of the metric as found in DataSourceBase.METRIC_MAP,
        db: database specified for this metric in grafana_client config
            options,
        attribute: the piece of information that will be selected from the
                   resource object to build the query.
        query: the unformatted query from the configuration for this metric,
        resource: the object from the OpenStackClient
        resource_type: the type of the resource
                       ['compute_node','instance', 'bare_metal', 'storage'],
        period: the period of time to collect metrics for in seconds,
        aggregate: the aggregation can be any from ['mean', 'max', 'min',
                   'count'],
        granularity: interval between datapoints in seconds (optional),
    }
    """

    """Every grafana translator should have a uniquely identifying name"""
    NAME = ''

    RESOURCE_TYPES = base.DataSourceBase.RESOURCE_TYPES

    AGGREGATES = base.DataSourceBase.AGGREGATES

    def __init__(self, data):
        self._data = data
        self._validate_data()

    def _validate_data(self):
        """iterate through the supplied data and verify attributes"""

        optionals = ['granularity']

        reference_data = {
            'metric': None,
            'db': None,
            'attribute': None,
            'query': None,
            'resource': None,
            'resource_type': None,
            'period': None,
            'aggregate': None,
            'granularity': None
        }
        reference_data.update(self._data)

        for key, value in reference_data.items():
            if value is None and key not in optionals:
                raise exception.InvalidParameter(
                    message=(_("The value %(value)s for parameter "
                               "%(parameter)s is invalid") % {'value': None,
                                                              'parameter': key}
                             )
                )

        if reference_data['resource_type'] not in self.RESOURCE_TYPES:
            raise exception.InvalidParameter(parameter='resource_type',
                                             parameter_type='RESOURCE_TYPES')

        if reference_data['aggregate'] not in self.AGGREGATES:
            raise exception.InvalidParameter(parameter='aggregate',
                                             parameter_type='AGGREGATES')

    @staticmethod
    def _extract_attribute(resource, attribute):
        """Retrieve the desired attribute from the resource

        :param resource: The resource object to extract the attribute from.
        :param attribute: The name of the attribute to subtract as string.
        :return: The extracted attribute or None
        """

        try:
            return getattr(resource, attribute)
        except AttributeError:
            raise

    @staticmethod
    def _query_format(query, aggregate, resource, period,
                      granularity, translator_specific):
        return query.format(aggregate, resource, period, granularity,
                            translator_specific)

    @abc.abstractmethod
    def build_params(self):
        """Build the set of parameters to send with the request"""
        raise NotImplementedError()

    @abc.abstractmethod
    def extract_result(self, raw_results):
        """Extrapolate the metric from the raw results of the request"""
        raise NotImplementedError()



"""ocata release

Revision ID: 9894235b4278
Revises: None
Create Date: 2017-02-01 09:40:05.065981

"""
from alembic import op
import oslo_db
import sqlalchemy as sa
from watcher.db.sqlalchemy import models


# revision identifiers, used by Alembic.
revision = '001'
down_revision = None


def upgrade():
    op.create_table(
        'goals',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('name', sa.String(length=63), nullable=False),
        sa.Column('display_name', sa.String(length=63), nullable=False),
        sa.Column('efficacy_specification', models.JSONEncodedList(),
                  nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'deleted', name='uniq_goals0name'),
        sa.UniqueConstraint('uuid', name='uniq_goals0uuid')
    )

    op.create_table(
        'scoring_engines',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=63), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('metainfo', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'deleted',
                            name='uniq_scoring_engines0name'),
        sa.UniqueConstraint('uuid', name='uniq_scoring_engines0uuid')
    )

    op.create_table(
        'services',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=False),
        sa.Column('last_seen_up', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('host', 'name', 'deleted',
                            name='uniq_services0host0name0deleted')
    )

    op.create_table(
        'strategies',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('name', sa.String(length=63), nullable=False),
        sa.Column('display_name', sa.String(length=63), nullable=False),
        sa.Column('goal_id', sa.Integer(), nullable=False),
        sa.Column('parameters_spec', models.JSONEncodedDict(),
                  nullable=True),
        sa.ForeignKeyConstraint(['goal_id'], ['goals.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'deleted', name='uniq_strategies0name'),
        sa.UniqueConstraint('uuid', name='uniq_strategies0uuid')
    )

    op.create_table(
        'audit_templates',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('name', sa.String(length=63), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('goal_id', sa.Integer(), nullable=False),
        sa.Column('strategy_id', sa.Integer(), nullable=True),
        sa.Column('scope', models.JSONEncodedList(),
                  nullable=True),
        sa.ForeignKeyConstraint(['goal_id'], ['goals.id'], ),
        sa.ForeignKeyConstraint(['strategy_id'], ['strategies.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'deleted',
                            name='uniq_audit_templates0name'),
        sa.UniqueConstraint('uuid', name='uniq_audit_templates0uuid')
    )
    op.create_table(
        'audits',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('audit_type', sa.String(length=20), nullable=True),
        sa.Column('state', sa.String(length=20), nullable=True),
        sa.Column('parameters', models.JSONEncodedDict(), nullable=True),
        sa.Column('interval', sa.Integer(), nullable=True),
        sa.Column('goal_id', sa.Integer(), nullable=False),
        sa.Column('strategy_id', sa.Integer(), nullable=True),
        sa.Column('scope', models.JSONEncodedList(), nullable=True),
        sa.Column('auto_trigger', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['goal_id'], ['goals.id'], ),
        sa.ForeignKeyConstraint(['strategy_id'], ['strategies.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_audits0uuid')
    )
    op.create_table(
        'action_plans',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('audit_id', sa.Integer(), nullable=False),
        sa.Column('strategy_id', sa.Integer(), nullable=False),
        sa.Column('state', sa.String(length=20), nullable=True),
        sa.Column('global_efficacy', models.JSONEncodedDict(), nullable=True),
        sa.ForeignKeyConstraint(['audit_id'], ['audits.id'], ),
        sa.ForeignKeyConstraint(['strategy_id'], ['strategies.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_action_plans0uuid')
    )

    op.create_table(
        'actions',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('action_plan_id', sa.Integer(), nullable=False),
        sa.Column('action_type', sa.String(length=255), nullable=False),
        sa.Column('input_parameters', models.JSONEncodedDict(), nullable=True),
        sa.Column('state', sa.String(length=20), nullable=True),
        sa.Column('parents', models.JSONEncodedList(), nullable=True),
        sa.ForeignKeyConstraint(['action_plan_id'], ['action_plans.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_actions0uuid')
    )

    op.create_table(
        'efficacy_indicators',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', oslo_db.sqlalchemy.types.SoftDeleteInteger(),
                  nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('name', sa.String(length=63), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('unit', sa.String(length=63), nullable=True),
        sa.Column('value', sa.Numeric(), nullable=True),
        sa.Column('action_plan_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['action_plan_id'], ['action_plans.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_efficacy_indicators0uuid')
    )


def downgrade():
    op.drop_table('efficacy_indicators')
    op.drop_table('actions')
    op.drop_table('action_plans')
    op.drop_table('audits')
    op.drop_table('audit_templates')
    op.drop_table('strategies')
    op.drop_table('services')
    op.drop_table('scoring_engines')
    op.drop_table('goals')



"""Set name for Audit as part of backward compatibility

Revision ID: a86240e89a29
Revises: 3cfc94cecf4e
Create Date: 2017-12-21 13:00:09.278587

"""

# revision identifiers, used by Alembic.
revision = 'a86240e89a29'
down_revision = '3cfc94cecf4e'

from alembic import op
from sqlalchemy.orm import sessionmaker
from watcher.db.sqlalchemy import models


def upgrade():
    connection = op.get_bind()
    session = sessionmaker()
    s = session(bind=connection)
    audits = s.query(
        models.Audit.strategy_id.label('strategy_id'),
        models.Audit.created_at.label('created_at')).filter(
        models.Audit.name is None).all()
    for audit in audits:
        strategy_name = s.query(models.Strategy).filter_by(
            id=audit.strategy_id).one().name
        s.query().filter(models.Audit.name is None).update(
            {'name': strategy_name + '-' + str(audit.created_at)})
    s.commit()


def downgrade():
    connection = op.get_bind()
    session = sessionmaker()
    s = session(bind=connection)
    audits = s.query(
        models.Audit.strategy_id.label('strategy_id'),
        models.Audit.created_at.label('created_at')).filter(
        models.Audit.name is not None).all()
    for audit in audits:
        s.query().filter(models.Audit.name is not None).update(
            {'name': None})
    s.commit()



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.decision_engine.goal import base as base_goal
from watcher.decision_engine.goal.efficacy import base as efficacy_base
from watcher.decision_engine.goal.efficacy import indicators
from watcher.decision_engine.goal.efficacy import specs


class FakeGoal(base_goal.Goal):

    NAME = NotImplemented
    DISPLAY_NAME = NotImplemented

    @classmethod
    def get_name(cls):
        return cls.NAME

    @classmethod
    def get_display_name(cls):
        return cls.DISPLAY_NAME

    @classmethod
    def get_translatable_display_name(cls):
        return cls.DISPLAY_NAME

    @classmethod
    def get_efficacy_specification(cls):
        """The efficacy spec for the current goal"""
        return specs.Unclassified()


class DummyIndicator(indicators.IndicatorSpecification):
    def __init__(self):
        super(DummyIndicator, self).__init__(
            name="dummy",
            description="Dummy indicator",
            unit="%",
        )

    @property
    def schema(self):
        return {
            "type": "integer",
            "minimum": 0
            }


class DummySpec1(efficacy_base.EfficacySpecification):

    def get_indicators_specifications(self):
        return [DummyIndicator()]

    def get_global_efficacy_indicator(self, indicators_map):
        return None


class FakeDummy1(FakeGoal):
    NAME = "dummy_1"
    DISPLAY_NAME = "Dummy 1"

    @classmethod
    def get_efficacy_specification(cls):
        """The efficacy spec for the current goal"""
        return DummySpec1()


class FakeDummy2(FakeGoal):
    NAME = "dummy_2"
    DISPLAY_NAME = "Dummy 2"



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Tests for manipulating Strategy via the DB API"""

import freezegun

from watcher.common import exception
from watcher.common import utils as w_utils
from watcher.tests.db import base
from watcher.tests.db import utils


class TestDbStrategyFilters(base.DbTestCase):

    FAKE_OLDER_DATE = '2014-01-01T09:52:05.219414'
    FAKE_OLD_DATE = '2015-01-01T09:52:05.219414'
    FAKE_TODAY = '2016-02-24T09:52:05.219414'

    def setUp(self):
        super(TestDbStrategyFilters, self).setUp()
        self.context.show_deleted = True
        self._data_setup()

    def _data_setup(self):
        strategy1_name = "STRATEGY_ID_1"
        strategy2_name = "STRATEGY_ID_2"
        strategy3_name = "STRATEGY_ID_3"

        self.goal1 = utils.create_test_goal(
            id=1, uuid=w_utils.generate_uuid(),
            name="GOAL_ID", display_name="Goal")
        self.goal2 = utils.create_test_goal(
            id=2, uuid=w_utils.generate_uuid(),
            name="DUMMY", display_name="Dummy")

        with freezegun.freeze_time(self.FAKE_TODAY):
            self.strategy1 = utils.create_test_strategy(
                id=1, uuid=w_utils.generate_uuid(),
                name=strategy1_name, display_name="Strategy 1",
                goal_id=self.goal1.id)
        with freezegun.freeze_time(self.FAKE_OLD_DATE):
            self.strategy2 = utils.create_test_strategy(
                id=2, uuid=w_utils.generate_uuid(),
                name=strategy2_name, display_name="Strategy 2",
                goal_id=self.goal1.id)
        with freezegun.freeze_time(self.FAKE_OLDER_DATE):
            self.strategy3 = utils.create_test_strategy(
                id=3, uuid=w_utils.generate_uuid(),
                name=strategy3_name, display_name="Strategy 3",
                goal_id=self.goal2.id)

    def _soft_delete_strategys(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.soft_delete_strategy(self.strategy1.id)
        with freezegun.freeze_time(self.FAKE_OLD_DATE):
            self.dbapi.soft_delete_strategy(self.strategy2.id)
        with freezegun.freeze_time(self.FAKE_OLDER_DATE):
            self.dbapi.soft_delete_strategy(self.strategy3.id)

    def _update_strategies(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.update_strategy(
                self.strategy1.id, values={"display_name": "strategy1"})
        with freezegun.freeze_time(self.FAKE_OLD_DATE):
            self.dbapi.update_strategy(
                self.strategy2.id, values={"display_name": "strategy2"})
        with freezegun.freeze_time(self.FAKE_OLDER_DATE):
            self.dbapi.update_strategy(
                self.strategy3.id, values={"display_name": "strategy3"})

    def test_get_strategy_list_filter_deleted_true(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.soft_delete_strategy(self.strategy1.id)

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted': True})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_deleted_false(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.soft_delete_strategy(self.strategy1.id)

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted': False})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_deleted_at_eq(self):
        self._soft_delete_strategys()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted_at__eq': self.FAKE_TODAY})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_deleted_at_lt(self):
        self._soft_delete_strategys()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted_at__lt': self.FAKE_TODAY})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_deleted_at_lte(self):
        self._soft_delete_strategys()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted_at__lte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_deleted_at_gt(self):
        self._soft_delete_strategys()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted_at__gt': self.FAKE_OLD_DATE})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_deleted_at_gte(self):
        self._soft_delete_strategys()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'deleted_at__gte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.strategy1['uuid'], self.strategy2['uuid']]),
            set([r.uuid for r in res]))

    # created_at #

    def test_get_strategy_list_filter_created_at_eq(self):
        res = self.dbapi.get_strategy_list(
            self.context, filters={'created_at__eq': self.FAKE_TODAY})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_created_at_lt(self):
        res = self.dbapi.get_strategy_list(
            self.context, filters={'created_at__lt': self.FAKE_TODAY})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_created_at_lte(self):
        res = self.dbapi.get_strategy_list(
            self.context, filters={'created_at__lte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_created_at_gt(self):
        res = self.dbapi.get_strategy_list(
            self.context, filters={'created_at__gt': self.FAKE_OLD_DATE})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_created_at_gte(self):
        res = self.dbapi.get_strategy_list(
            self.context, filters={'created_at__gte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.strategy1['uuid'], self.strategy2['uuid']]),
            set([r.uuid for r in res]))

    # updated_at #

    def test_get_strategy_list_filter_updated_at_eq(self):
        self._update_strategies()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'updated_at__eq': self.FAKE_TODAY})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_updated_at_lt(self):
        self._update_strategies()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'updated_at__lt': self.FAKE_TODAY})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_updated_at_lte(self):
        self._update_strategies()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'updated_at__lte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.strategy2['uuid'], self.strategy3['uuid']]),
            set([r.uuid for r in res]))

    def test_get_strategy_list_filter_updated_at_gt(self):
        self._update_strategies()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'updated_at__gt': self.FAKE_OLD_DATE})

        self.assertEqual([self.strategy1['uuid']], [r.uuid for r in res])

    def test_get_strategy_list_filter_updated_at_gte(self):
        self._update_strategies()

        res = self.dbapi.get_strategy_list(
            self.context, filters={'updated_at__gte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.strategy1['uuid'], self.strategy2['uuid']]),
            set([r.uuid for r in res]))


class DbStrategyTestCase(base.DbTestCase):

    def test_get_strategy_list(self):
        uuids = []
        for i in range(1, 4):
            strategy = utils.create_test_strategy(
                id=i,
                uuid=w_utils.generate_uuid(),
                name="STRATEGY_ID_%s" % i,
                display_name='My Strategy {0}'.format(i))
            uuids.append(str(strategy['uuid']))
        strategies = self.dbapi.get_strategy_list(self.context)
        strategy_uuids = [s.uuid for s in strategies]
        self.assertEqual(sorted(uuids), sorted(strategy_uuids))
        for strategy in strategies:
            self.assertIsNone(strategy.goal)

    def test_get_strategy_list_eager(self):
        _goal = utils.get_test_goal()
        goal = self.dbapi.create_goal(_goal)
        uuids = []
        for i in range(1, 4):
            strategy = utils.create_test_strategy(
                id=i,
                uuid=w_utils.generate_uuid(),
                name="STRATEGY_ID_%s" % i,
                display_name='My Strategy {0}'.format(i),
                goal_id=goal.id)
            uuids.append(str(strategy['uuid']))
        strategys = self.dbapi.get_strategy_list(self.context, eager=True)
        strategy_map = {a.uuid: a for a in strategys}
        self.assertEqual(sorted(uuids), sorted(strategy_map.keys()))
        eager_strategy = strategy_map[strategy.uuid]
        self.assertEqual(goal.as_dict(), eager_strategy.goal.as_dict())

    def test_get_strategy_list_with_filters(self):
        # NOTE(erakli): we don't create goal in database but links to
        # goal_id = 1. There is no error in dbapi.create_strategy() method.
        # Is it right behaviour?
        strategy1 = utils.create_test_strategy(
            id=1,
            uuid=w_utils.generate_uuid(),
            name="STRATEGY_ID_1",
            display_name='Strategy 1',
        )
        strategy2 = utils.create_test_strategy(
            id=2,
            uuid=w_utils.generate_uuid(),
            name="STRATEGY_ID_2",
            display_name='Strategy 2',
        )
        strategy3 = utils.create_test_strategy(
            id=3,
            uuid=w_utils.generate_uuid(),
            name="STRATEGY_ID_3",
            display_name='Strategy 3',
        )

        self.dbapi.soft_delete_strategy(strategy3['uuid'])

        res = self.dbapi.get_strategy_list(
            self.context, filters={'display_name': 'Strategy 1'})
        self.assertEqual([strategy1['uuid']], [r.uuid for r in res])

        res = self.dbapi.get_strategy_list(
            self.context, filters={'display_name': 'Strategy 3'})
        self.assertEqual([], [r.uuid for r in res])

        res = self.dbapi.get_strategy_list(
            self.context, filters={'goal_id': 1})
        self.assertEqual([strategy1['uuid'], strategy2['uuid']],
                         [r.uuid for r in res])

        res = self.dbapi.get_strategy_list(
            self.context, filters={'display_name': 'Strategy 2'})
        self.assertEqual([strategy2['uuid']], [r.uuid for r in res])

    def test_get_strategy_by_uuid(self):
        created_strategy = utils.create_test_strategy()
        strategy = self.dbapi.get_strategy_by_uuid(
            self.context, created_strategy['uuid'])
        self.assertEqual(strategy.uuid, created_strategy['uuid'])

    def test_get_strategy_by_name(self):
        created_strategy = utils.create_test_strategy()
        strategy = self.dbapi.get_strategy_by_name(
            self.context, created_strategy['name'])
        self.assertEqual(strategy.name, created_strategy['name'])

    def test_get_strategy_that_does_not_exist(self):
        self.assertRaises(exception.StrategyNotFound,
                          self.dbapi.get_strategy_by_id,
                          self.context, 404)

    def test_update_strategy(self):
        strategy = utils.create_test_strategy()
        res = self.dbapi.update_strategy(
            strategy['uuid'], {'display_name': 'updated-model'})
        self.assertEqual('updated-model', res.display_name)

    def test_update_goal_id(self):
        strategy = utils.create_test_strategy()
        self.assertRaises(exception.Invalid,
                          self.dbapi.update_strategy, strategy['uuid'],
                          {'uuid': 'new_strategy_id'})

    def test_update_strategy_that_does_not_exist(self):
        self.assertRaises(exception.StrategyNotFound,
                          self.dbapi.update_strategy,
                          404,
                          {'display_name': ''})

    def test_destroy_strategy(self):
        strategy = utils.create_test_strategy()
        self.dbapi.destroy_strategy(strategy['uuid'])
        self.assertRaises(exception.StrategyNotFound,
                          self.dbapi.get_strategy_by_id,
                          self.context, strategy['uuid'])

    def test_destroy_strategy_that_does_not_exist(self):
        self.assertRaises(exception.StrategyNotFound,
                          self.dbapi.destroy_strategy, 404)

    def test_create_strategy_already_exists(self):
        strategy_id = "STRATEGY_ID"
        utils.create_test_strategy(name=strategy_id)
        self.assertRaises(exception.StrategyAlreadyExists,
                          utils.create_test_strategy,
                          name=strategy_id)



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from watcher._i18n import _
from watcher.decision_engine.strategy.strategies import base


class Actuator(base.UnclassifiedStrategy):
    """Actuator

    Actuator that simply executes the actions given as parameter

    This strategy allows anyone to create an action plan with a predefined
    set of actions. This strategy can be used for 2 different purposes:

    - Test actions
    - Use this strategy based on an event trigger to perform some explicit task

    """

    @classmethod
    def get_name(cls):
        return "actuator"

    @classmethod
    def get_display_name(cls):
        return _("Actuator")

    @classmethod
    def get_translatable_display_name(cls):
        return "Actuator"

    @classmethod
    def get_schema(cls):
        # Mandatory default setting for each element
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "actions": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "action_type": {
                                "type": "string"
                            },
                            "resource_id": {
                                "type": "string"
                            },
                            "input_parameters": {
                                "type": "object",
                                "properties": {},
                                "additionalProperties": True
                            }
                        },
                        "required": [
                            "action_type", "input_parameters"
                        ],
                        "additionalProperties": True,
                    }
                }
            },
            "required": [
                "actions"
            ]
        }

    @classmethod
    def get_config_opts(cls):
        """Override base class config options as do not use datasource """

        return []

    @property
    def actions(self):
        return self.input_parameters.get('actions', [])

    def pre_execute(self):
        self._pre_execute()

    def do_execute(self, audit=None):
        for action in self.actions:
            self.solution.add_action(**action)

    def post_execute(self):
        pass



# -*- encoding: utf-8 -*-
#
# Authors: Vincent Francoise <Vincent.FRANCOISE@b-com.com>
#          Alexander Chadin <a.chadin@servionica.ru>
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

import networkx as nx
from oslo_config import cfg
from oslo_log import log

from watcher.common import utils
from watcher.decision_engine.planner import base
from watcher import objects

LOG = log.getLogger(__name__)


class WeightPlanner(base.BasePlanner):
    """Weight planner implementation

    This implementation builds actions with parents in accordance with weights.
    Set of actions having a higher weight will be scheduled before
    the other ones. There are two config options to configure:
    action_weights and parallelization.

    *Limitations*

    - This planner requires to have action_weights and parallelization configs
      tuned well.
    """

    def __init__(self, config):
        super(WeightPlanner, self).__init__(config)

    action_weights = {
        'nop': 70,
        'volume_migrate': 60,
        'change_nova_service_state': 50,
        'sleep': 40,
        'migrate': 30,
        'resize': 20,
        'turn_host_to_acpi_s3_state': 10,
        'change_node_power_state': 9,
    }

    parallelization = {
        'turn_host_to_acpi_s3_state': 2,
        'resize': 2,
        'migrate': 2,
        'sleep': 1,
        'change_nova_service_state': 1,
        'nop': 1,
        'change_node_power_state': 2,
        'volume_migrate': 2
    }

    @classmethod
    def get_config_opts(cls):
        return [
            cfg.DictOpt(
                'weights',
                help="These weights are used to schedule the actions. "
                     "Action Plan will be build in accordance with sets of "
                     "actions ordered by descending weights."
                     "Two action types cannot have the same weight. ",
                default=cls.action_weights),
            cfg.DictOpt(
                'parallelization',
                help="Number of actions to be run in parallel on a per "
                     "action type basis.",
                default=cls.parallelization),
        ]

    @staticmethod
    def chunkify(lst, n):
        """Yield successive n-sized chunks from lst."""
        n = int(n)
        if n < 1:
            # Just to make sure the number is valid
            n = 1

        # Split a flat list in a list of chunks of size n.
        # e.g. chunkify([0, 1, 2, 3, 4], 2) -> [[0, 1], [2, 3], [4]]
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def compute_action_graph(self, sorted_weighted_actions):
        reverse_weights = {v: k for k, v in self.config.weights.items()}
        # leaf_groups contains a list of list of nodes called groups
        # each group is a set of nodes from which a future node will
        # branch off (parent nodes).

        # START --> migrate-1 --> migrate-3
        #     \                            \--> resize-1 --> FINISH
        #      \--> migrate-2 -------------/
        # In the above case migrate-1 will be the only member of the leaf
        # group that migrate-3 will use as parent group, whereas
        # resize-1 will have both migrate-2 and migrate-3 in its
        # parent/leaf group
        leaf_groups = []
        action_graph = nx.DiGraph()
        # We iterate through each action type category (sorted by weight) to
        # insert them in a Directed Acyclic Graph
        for idx, (weight, actions) in enumerate(sorted_weighted_actions):
            action_chunks = self.chunkify(
                actions, self.config.parallelization[reverse_weights[weight]])

            # We split the actions into chunks/layers that will have to be
            # spread across all the available branches of the graph
            for chunk_idx, actions_chunk in enumerate(action_chunks):
                for action in actions_chunk:
                    action_graph.add_node(action)

                    # all other actions
                    parent_nodes = []
                    if not idx and not chunk_idx:
                        parent_nodes = []
                    elif leaf_groups:
                        parent_nodes = leaf_groups

                    for parent_node in parent_nodes:
                        action_graph.add_edge(parent_node, action)
                        action.parents.append(parent_node.uuid)

                if leaf_groups:
                    leaf_groups = []
                leaf_groups.extend([a for a in actions_chunk])

        return action_graph

    def schedule(self, context, audit_id, solution):
        LOG.debug('Creating an action plan for the audit uuid: %s', audit_id)
        action_plan = self.create_action_plan(context, audit_id, solution)

        sorted_weighted_actions = self.get_sorted_actions_by_weight(
            context, action_plan, solution)
        action_graph = self.compute_action_graph(sorted_weighted_actions)

        self._create_efficacy_indicators(
            context, action_plan.id, solution.efficacy_indicators)

        if len(action_graph.nodes()) == 0:
            LOG.warning("The action plan is empty")
            action_plan.state = objects.action_plan.State.SUCCEEDED
            action_plan.save()

        self.create_scheduled_actions(action_graph)
        return action_plan

    def get_sorted_actions_by_weight(self, context, action_plan, solution):
        # We need to make them immutable to add them to the graph
        action_objects = list([
            objects.Action(
                context, uuid=utils.generate_uuid(), parents=[],
                action_plan_id=action_plan.id, **a)
            for a in solution.actions])
        # This is a dict of list with each being a weight and the list being
        # all the actions associated to this weight
        weighted_actions = collections.defaultdict(list)
        for action in action_objects:
            action_weight = self.config.weights[action.action_type]
            weighted_actions[action_weight].append(action)

        return reversed(sorted(weighted_actions.items(), key=lambda x: x[0]))

    def create_scheduled_actions(self, graph):
        for action in graph.nodes():
            LOG.debug("Creating the %s in the Watcher database",
                      action.action_type)
            try:
                action.create()
            except Exception as exc:
                LOG.exception(exc)
                raise

    def create_action_plan(self, context, audit_id, solution):
        strategy = objects.Strategy.get_by_name(
            context, solution.strategy.name)

        action_plan_dict = {
            'uuid': utils.generate_uuid(),
            'audit_id': audit_id,
            'strategy_id': strategy.id,
            'state': objects.action_plan.State.RECOMMENDED,
            'global_efficacy': solution.global_efficacy,
        }

        new_action_plan = objects.ActionPlan(context, **action_plan_dict)
        new_action_plan.create()

        return new_action_plan

    def _create_efficacy_indicators(self, context, action_plan_id, indicators):
        efficacy_indicators = []
        for indicator in indicators:
            efficacy_indicator_dict = {
                'uuid': utils.generate_uuid(),
                'name': indicator.name,
                'description': indicator.description,
                'unit': indicator.unit,
                'value': indicator.value,
                'action_plan_id': action_plan_id,
            }
            new_efficacy_indicator = objects.EfficacyIndicator(
                context, **efficacy_indicator_dict)
            new_efficacy_indicator.create()

            efficacy_indicators.append(new_efficacy_indicator)
        return efficacy_indicators



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#          Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from watcher import notifications
from watcher.objects import fields


class StrategyContext(object, metaclass=abc.ABCMeta):

    def execute_strategy(self, audit, request_context):
        """Execute the strategy for the given an audit

        :param audit: Audit object
        :type audit: :py:class:`~.objects.audit.Audit` instance
        :param request_context: Current request context
        :type request_context: :py:class:`~.RequestContext` instance
        :returns: The computed solution
        :rtype: :py:class:`~.BaseSolution` instance
        """
        try:
            notifications.audit.send_action_notification(
                request_context, audit,
                action=fields.NotificationAction.STRATEGY,
                phase=fields.NotificationPhase.START)
            solution = self.do_execute_strategy(audit, request_context)
            notifications.audit.send_action_notification(
                request_context, audit,
                action=fields.NotificationAction.STRATEGY,
                phase=fields.NotificationPhase.END)
            return solution
        except Exception:
            notifications.audit.send_action_notification(
                request_context, audit,
                action=fields.NotificationAction.STRATEGY,
                priority=fields.NotificationPriority.ERROR,
                phase=fields.NotificationPhase.ERROR)
            raise

    @abc.abstractmethod
    def do_execute_strategy(self, audit, request_context):
        """Execute the strategy for the given an audit

        :param audit: Audit object
        :type audit: :py:class:`~.objects.audit.Audit` instance
        :param request_context: Current request context
        :type request_context: :py:class:`~.RequestContext` instance
        :returns: The computed solution
        :rtype: :py:class:`~.BaseSolution` instance
        """
        raise NotImplementedError()



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel Corp
#
# Authors: Junjie-Huang <junjie.huang@intel.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import collections
from unittest import mock

from watcher.applier.loading import default
from watcher.common import utils
from watcher.decision_engine.strategy import strategies
from watcher.tests.decision_engine.model import ceilometer_metrics
from watcher.tests.decision_engine.model import gnocchi_metrics
from watcher.tests.decision_engine.strategy.strategies.test_base \
    import TestBaseStrategy


class TestUniformAirflow(TestBaseStrategy):

    scenarios = [
        ("Ceilometer",
         {"datasource": "ceilometer",
          "fake_datasource_cls": ceilometer_metrics.FakeCeilometerMetrics}),
        ("Gnocchi",
         {"datasource": "gnocchi",
          "fake_datasource_cls": gnocchi_metrics.FakeGnocchiMetrics}),
    ]

    def setUp(self):
        super(TestUniformAirflow, self).setUp()
        # fake metrics
        self.fake_metrics = self.fake_datasource_cls()

        p_datasource = mock.patch.object(
            strategies.UniformAirflow, 'datasource_backend',
            new_callable=mock.PropertyMock)
        self.m_datasource = p_datasource.start()
        self.addCleanup(p_datasource.stop)

        self.m_datasource.return_value = mock.Mock(
            statistic_aggregation=self.fake_metrics.mock_get_statistics,
            NAME=self.fake_metrics.NAME)
        self.strategy = strategies.UniformAirflow(
            config=mock.Mock(datasource=self.datasource))
        self.strategy.input_parameters = utils.Struct()
        self.strategy.input_parameters.update({'threshold_airflow': 400.0,
                                               'threshold_inlet_t': 28.0,
                                               'threshold_power': 350.0,
                                               'period': 300})
        self.strategy.threshold_airflow = 400
        self.strategy.threshold_inlet_t = 28
        self.strategy.threshold_power = 350
        self._period = 300
        self.strategy.pre_execute()

    def test_calc_used_resource(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        node = model.get_node_by_uuid('Node_0')
        cores_used, mem_used, disk_used = (
            self.strategy.calculate_used_resource(node))
        self.assertEqual((cores_used, mem_used, disk_used), (25, 4, 40))

    def test_group_hosts_by_airflow(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        self.strategy.threshold_airflow = 300
        n1, n2 = self.strategy.group_hosts_by_airflow()
        # print n1, n2, avg, w_map
        self.assertEqual(n1[0]['node'].uuid, 'Node_0')
        self.assertEqual(n2[0]['node'].uuid, 'Node_1')

    def test_choose_instance_to_migrate(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        self.strategy.threshold_airflow = 300
        self.strategy.threshold_inlet_t = 22
        n1, n2 = self.strategy.group_hosts_by_airflow()
        instance_to_mig = self.strategy.choose_instance_to_migrate(n1)

        self.assertEqual(instance_to_mig[0].uuid, 'Node_0')
        self.assertEqual(len(instance_to_mig[1]), 1)
        self.assertIn(instance_to_mig[1][0].uuid,
                      {'cae81432-1631-4d4e-b29c-6f3acdcde906',
                       '73b09e16-35b7-4922-804e-e8f5d9b740fc'})

    def test_choose_instance_to_migrate_all(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        self.strategy.threshold_airflow = 300
        self.strategy.threshold_inlet_t = 25
        n1, n2 = self.strategy.group_hosts_by_airflow()
        instance_to_mig = self.strategy.choose_instance_to_migrate(n1)

        self.assertEqual(instance_to_mig[0].uuid, 'Node_0')
        self.assertEqual(len(instance_to_mig[1]), 2)
        self.assertEqual({'cae81432-1631-4d4e-b29c-6f3acdcde906',
                          '73b09e16-35b7-4922-804e-e8f5d9b740fc'},
                         {inst.uuid for inst in instance_to_mig[1]})

    def test_choose_instance_notfound(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        self.strategy.threshold_airflow = 300
        self.strategy.threshold_inlet_t = 22
        n1, n2 = self.strategy.group_hosts_by_airflow()
        instances = model.get_all_instances()
        [model.remove_instance(inst) for inst in instances.values()]
        instance_to_mig = self.strategy.choose_instance_to_migrate(n1)
        self.assertIsNone(instance_to_mig)

    def test_filter_destination_hosts(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        self.strategy.threshold_airflow = 300
        self.strategy.threshold_inlet_t = 22
        n1, n2 = self.strategy.group_hosts_by_airflow()
        instance_to_mig = self.strategy.choose_instance_to_migrate(n1)
        dest_hosts = self.strategy.filter_destination_hosts(
            n2, instance_to_mig[1])

        self.assertEqual(len(dest_hosts), 1)
        self.assertEqual(dest_hosts[0]['node'].uuid, 'Node_1')
        self.assertIn(instance_to_mig[1][0].uuid,
                      {'cae81432-1631-4d4e-b29c-6f3acdcde906',
                       '73b09e16-35b7-4922-804e-e8f5d9b740fc'})

    def test_execute_no_workload(self):
        self.strategy.threshold_airflow = 300
        self.strategy.threshold_inlet_t = 25
        self.strategy.threshold_power = 300
        model = self.fake_c_cluster.\
            generate_scenario_4_with_1_node_no_instance()
        self.m_c_model.return_value = model
        solution = self.strategy.execute()
        self.assertEqual([], solution.actions)

    def test_execute(self):
        self.strategy.threshold_airflow = 300
        self.strategy.threshold_inlet_t = 25
        self.strategy.threshold_power = 300
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        solution = self.strategy.execute()
        actions_counter = collections.Counter(
            [action.get('action_type') for action in solution.actions])

        num_migrations = actions_counter.get("migrate", 0)
        self.assertEqual(num_migrations, 2)

    def test_check_parameters(self):
        model = self.fake_c_cluster.generate_scenario_7_with_2_nodes()
        self.m_c_model.return_value = model
        solution = self.strategy.execute()
        loader = default.DefaultActionLoader()
        for action in solution.actions:
            loaded_action = loader.load(action['action_type'])
            loaded_action.input_parameters = action['input_parameters']
            loaded_action.validate_parameters()



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Authors: Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher.decision_engine import gmr
from watcher.decision_engine.model.collector import manager
from watcher.tests import base


class TestGmrPlugin(base.TestCase):

    @mock.patch.object(manager.CollectorManager, "get_collectors")
    def test_show_models(self, m_get_collectors):
        m_to_string = mock.Mock(return_value="<TESTMODEL />")
        m_get_collectors.return_value = {
            "test_model": mock.Mock(
                cluster_data_model=mock.Mock(to_string=m_to_string))}
        output = gmr.show_models()
        self.assertEqual(1, m_to_string.call_count)
        self.assertIn("<TESTMODEL />", output)



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stevedore import extension
from unittest import mock

from watcher.common import exception
from watcher.decision_engine.loading import default as default_loading
from watcher.decision_engine.strategy.strategies import dummy_strategy
from watcher.tests import base


class TestDefaultStrategyLoader(base.TestCase):

    def setUp(self):
        super(TestDefaultStrategyLoader, self).setUp()
        self.strategy_loader = default_loading.DefaultStrategyLoader()

    def test_load_strategy_with_empty_model(self):
        self.assertRaises(
            exception.LoadingError, self.strategy_loader.load, None)

    def test_strategy_loader(self):
        dummy_strategy_name = "dummy"
        # Set up the fake Stevedore extensions
        fake_extmanager_call = extension.ExtensionManager.make_test_instance(
            extensions=[extension.Extension(
                name=dummy_strategy_name,
                entry_point="%s:%s" % (
                    dummy_strategy.DummyStrategy.__module__,
                    dummy_strategy.DummyStrategy.__name__),
                plugin=dummy_strategy.DummyStrategy,
                obj=None,
            )],
            namespace="watcher_strategies",
        )

        with mock.patch.object(extension, "ExtensionManager") as m_ext_manager:
            m_ext_manager.return_value = fake_extmanager_call
            loaded_strategy = self.strategy_loader.load(
                "dummy")

        self.assertEqual("dummy", loaded_strategy.name)
        self.assertEqual("Dummy strategy", loaded_strategy.display_name)

    def test_load_dummy_strategy(self):
        strategy_loader = default_loading.DefaultStrategyLoader()
        loaded_strategy = strategy_loader.load("dummy")
        self.assertIsInstance(loaded_strategy, dummy_strategy.DummyStrategy)


class TestLoadStrategiesWithDefaultStrategyLoader(base.TestCase):

    strategy_loader = default_loading.DefaultStrategyLoader()

    scenarios = [
        (strategy_name,
         {"strategy_name": strategy_name, "strategy_cls": strategy_cls})
        for strategy_name, strategy_cls
        in strategy_loader.list_available().items()]

    def test_load_strategies(self):
        strategy = self.strategy_loader.load(self.strategy_name)
        self.assertIsNotNone(strategy)
        self.assertEqual(self.strategy_name, strategy.name)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

import iso8601

from watcher.common import exception
from watcher.common import utils as c_utils
from watcher.db.sqlalchemy import api as db_api
from watcher import notifications
from watcher import objects
from watcher.tests.db import base
from watcher.tests.db import utils


class TestActionObject(base.DbTestCase):

    action_plan_id = 2

    scenarios = [
        ('non_eager', dict(
            eager=False,
            fake_action=utils.get_test_action(
                action_plan_id=action_plan_id))),
        ('eager_with_non_eager_load', dict(
            eager=True,
            fake_action=utils.get_test_action(
                action_plan_id=action_plan_id))),
        ('eager_with_eager_load', dict(
            eager=True,
            fake_action=utils.get_test_action(
                action_plan_id=action_plan_id,
                action_plan=utils.get_test_action_plan(id=action_plan_id)))),
    ]

    def setUp(self):
        super(TestActionObject, self).setUp()

        p_action_notifications = mock.patch.object(
            notifications, 'action_plan', autospec=True)
        self.m_action_notifications = p_action_notifications.start()
        self.addCleanup(p_action_notifications.stop)
        self.m_send_update = self.m_action_notifications.send_update

        self.fake_action_plan = utils.create_test_action_plan(
            id=self.action_plan_id)

    def eager_action_assert(self, action):
        if self.eager:
            self.assertIsNotNone(action.action_plan)
            fields_to_check = set(
                super(objects.ActionPlan, objects.ActionPlan).fields
            ).symmetric_difference(objects.ActionPlan.fields)
            db_data = {
                k: v for k, v in self.fake_action_plan.as_dict().items()
                if k in fields_to_check}
            object_data = {
                k: v for k, v in action.action_plan.as_dict().items()
                if k in fields_to_check}
            self.assertEqual(db_data, object_data)

    @mock.patch.object(db_api.Connection, 'get_action_by_id')
    def test_get_by_id(self, mock_get_action):
        mock_get_action.return_value = self.fake_action
        action_id = self.fake_action['id']
        action = objects.Action.get(self.context, action_id, eager=self.eager)
        mock_get_action.assert_called_once_with(
            self.context, action_id, eager=self.eager)
        self.assertEqual(self.context, action._context)
        self.eager_action_assert(action)
        self.assertEqual(0, self.m_send_update.call_count)

    @mock.patch.object(db_api.Connection, 'get_action_by_uuid')
    def test_get_by_uuid(self, mock_get_action):
        mock_get_action.return_value = self.fake_action
        uuid = self.fake_action['uuid']
        action = objects.Action.get(self.context, uuid, eager=self.eager)
        mock_get_action.assert_called_once_with(
            self.context, uuid, eager=self.eager)
        self.assertEqual(self.context, action._context)
        self.assertEqual(0, self.m_send_update.call_count)

    def test_get_bad_id_and_uuid(self):
        self.assertRaises(exception.InvalidIdentity,
                          objects.Action.get, self.context, 'not-a-uuid',
                          eager=self.eager)

    @mock.patch.object(db_api.Connection, 'get_action_list')
    def test_list(self, mock_get_list):
        mock_get_list.return_value = [self.fake_action]
        actions = objects.Action.list(self.context, eager=self.eager)
        self.assertEqual(1, mock_get_list.call_count)
        self.assertEqual(1, len(actions))
        self.assertIsInstance(actions[0], objects.Action)
        self.assertEqual(self.context, actions[0]._context)
        for action in actions:
            self.eager_action_assert(action)
        self.assertEqual(0, self.m_send_update.call_count)

    @mock.patch.object(objects.Strategy, 'get')
    @mock.patch.object(objects.Audit, 'get')
    @mock.patch.object(db_api.Connection, 'update_action')
    @mock.patch.object(db_api.Connection, 'get_action_by_uuid')
    def test_save(self, mock_get_action, mock_update_action, mock_get_audit,
                  mock_get_strategy):
        mock_get_action.return_value = self.fake_action
        fake_saved_action = self.fake_action.copy()
        mock_get_audit.return_value = mock.PropertyMock(
            uuid=c_utils.generate_uuid())
        mock_get_strategy.return_value = mock.PropertyMock(
            uuid=c_utils.generate_uuid())
        fake_saved_action['updated_at'] = datetime.datetime.utcnow()
        mock_update_action.return_value = fake_saved_action
        uuid = self.fake_action['uuid']
        action = objects.Action.get_by_uuid(
            self.context, uuid, eager=self.eager)
        action.state = objects.action.State.SUCCEEDED
        if not self.eager:
            self.assertRaises(exception.EagerlyLoadedActionRequired,
                              action.save)
        else:
            action.save()

        expected_update_at = fake_saved_action['updated_at'].replace(
            tzinfo=iso8601.UTC)

        mock_get_action.assert_called_once_with(
            self.context, uuid, eager=self.eager)
        mock_update_action.assert_called_once_with(
            uuid, {'state': objects.action.State.SUCCEEDED})
        self.assertEqual(self.context, action._context)
        self.assertEqual(expected_update_at, action.updated_at)
        self.assertEqual(0, self.m_send_update.call_count)

    @mock.patch.object(db_api.Connection, 'get_action_by_uuid')
    def test_refresh(self, mock_get_action):
        returns = [dict(self.fake_action, state="first state"),
                   dict(self.fake_action, state="second state")]
        mock_get_action.side_effect = returns
        uuid = self.fake_action['uuid']
        expected = [mock.call(self.context, uuid, eager=self.eager),
                    mock.call(self.context, uuid, eager=self.eager)]
        action = objects.Action.get(self.context, uuid, eager=self.eager)
        self.assertEqual("first state", action.state)
        action.refresh(eager=self.eager)
        self.assertEqual("second state", action.state)
        self.assertEqual(expected, mock_get_action.call_args_list)
        self.assertEqual(self.context, action._context)
        self.eager_action_assert(action)
        self.assertEqual(0, self.m_send_update.call_count)


class TestCreateDeleteActionObject(base.DbTestCase):

    def setUp(self):
        super(TestCreateDeleteActionObject, self).setUp()
        self.fake_strategy = utils.create_test_strategy(name="DUMMY")
        self.fake_audit = utils.create_test_audit()
        self.fake_action_plan = utils.create_test_action_plan()
        self.fake_action = utils.get_test_action(
            created_at=datetime.datetime.utcnow())

    @mock.patch.object(db_api.Connection, 'create_action')
    def test_create(self, mock_create_action):
        mock_create_action.return_value = self.fake_action
        action = objects.Action(self.context, **self.fake_action)
        action.create()
        expected_action = self.fake_action.copy()
        expected_action['created_at'] = expected_action['created_at'].replace(
            tzinfo=iso8601.UTC)
        mock_create_action.assert_called_once_with(expected_action)
        self.assertEqual(self.context, action._context)

    @mock.patch.object(notifications.action, 'send_delete')
    @mock.patch.object(notifications.action, 'send_update')
    @mock.patch.object(db_api.Connection, 'update_action')
    @mock.patch.object(db_api.Connection, 'soft_delete_action')
    @mock.patch.object(db_api.Connection, 'get_action_by_uuid')
    def test_soft_delete(self, mock_get_action,
                         mock_soft_delete_action, mock_update_action,
                         mock_send_update, mock_send_delete):
        mock_get_action.return_value = self.fake_action
        fake_deleted_action = self.fake_action.copy()
        fake_deleted_action['deleted_at'] = datetime.datetime.utcnow()
        mock_soft_delete_action.return_value = fake_deleted_action
        mock_update_action.return_value = fake_deleted_action

        expected_action = fake_deleted_action.copy()
        expected_action['created_at'] = expected_action['created_at'].replace(
            tzinfo=iso8601.UTC)
        expected_action['deleted_at'] = expected_action['deleted_at'].replace(
            tzinfo=iso8601.UTC)
        del expected_action['action_plan']

        uuid = self.fake_action['uuid']
        action = objects.Action.get_by_uuid(self.context, uuid)
        action.soft_delete()
        mock_get_action.assert_called_once_with(
            self.context, uuid, eager=False)
        mock_soft_delete_action.assert_called_once_with(uuid)
        mock_update_action.assert_called_once_with(
            uuid, {'state': objects.action.State.DELETED})
        self.assertEqual(self.context, action._context)
        self.assertEqual(expected_action, action.as_dict())

    @mock.patch.object(db_api.Connection, 'destroy_action')
    @mock.patch.object(db_api.Connection, 'get_action_by_uuid')
    def test_destroy(self, mock_get_action, mock_destroy_action):
        mock_get_action.return_value = self.fake_action
        uuid = self.fake_action['uuid']
        action = objects.Action.get_by_uuid(self.context, uuid)
        action.destroy()

        mock_get_action.assert_called_once_with(
            self.context, uuid, eager=False)
        mock_destroy_action.assert_called_once_with(uuid)
        self.assertEqual(self.context, action._context)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import mock

from http import HTTPStatus
import time

from cinderclient import exceptions as cinder_exception

from watcher.common import cinder_helper
from watcher.common import clients
from watcher.common import exception
from watcher.common import utils
from watcher.tests import base


@mock.patch.object(clients.OpenStackClients, 'cinder')
class TestCinderHelper(base.TestCase):

    @staticmethod
    def fake_storage_node(**kwargs):
        node = mock.MagicMock()
        node.binary = kwargs.get('binary', 'cinder-volume')
        node.host = kwargs.get('name', 'host@backend')

        return node

    def test_get_storage_node_list(self, mock_cinder):
        node1 = self.fake_storage_node()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.services.list.return_value = [node1]
        cinder_util.get_storage_node_list()
        cinder_util.cinder.services.list.assert_called_once_with(
            binary='cinder-volume')

    def test_get_storage_node_by_name_success(self, mock_cinder):
        node1 = self.fake_storage_node()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.services.list.return_value = [node1]
        node = cinder_util.get_storage_node_by_name('host@backend')

        self.assertEqual(node, node1)

    def test_get_storage_node_by_name_failure(self, mock_cinder):
        node1 = self.fake_storage_node()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.services.list.return_value = [node1]
        self.assertRaisesRegex(
            exception.StorageNodeNotFound,
            "The storage node failure could not be found",
            cinder_util.get_storage_node_by_name, 'failure')

    @staticmethod
    def fake_pool(**kwargs):
        pool = mock.MagicMock()
        pool.name = kwargs.get('name', 'host@backend#pool')

        return pool

    def test_get_storage_pool_list(self, mock_cinder):
        pool = self.fake_pool()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.pools.list.return_value = [pool]
        cinder_util.get_storage_pool_list()
        cinder_util.cinder.pools.list.assert_called_once_with(detailed=True)

    def test_get_storage_pool_by_name_success(self, mock_cinder):
        pool1 = self.fake_pool()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.pools.list.return_value = [pool1]
        pool = cinder_util.get_storage_pool_by_name('host@backend#pool')

        self.assertEqual(pool, pool1)

    def test_get_storage_pool_by_name_failure(self, mock_cinder):
        pool1 = self.fake_pool()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.services.list.return_value = [pool1]
        self.assertRaisesRegex(
            exception.PoolNotFound,
            "The pool failure could not be found",
            cinder_util.get_storage_pool_by_name, 'failure')

    @staticmethod
    def fake_volume_type(**kwargs):
        volume_type = mock.MagicMock()
        volume_type.name = kwargs.get('name', 'fake_type')
        extra_specs = {'volume_backend_name': 'backend'}
        volume_type.extra_specs = kwargs.get('extra_specs', extra_specs)
        return volume_type

    def test_get_volume_type_list(self, mock_cinder):
        volume_type1 = self.fake_volume_type()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.volume_types.list.return_value = [volume_type1]
        cinder_util.get_volume_type_list()
        cinder_util.cinder.volume_types.list.assert_called_once_with()

    def test_get_volume_type_by_backendname_with_backend_exist(
            self, mock_cinder):
        volume_type1 = self.fake_volume_type()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.volume_types.list.return_value = [volume_type1]
        volume_type_name = cinder_util.get_volume_type_by_backendname(
            'backend')

        self.assertEqual(volume_type_name[0], volume_type1.name)

    def test_get_volume_type_by_backendname_with_no_backend_exist(
            self, mock_cinder):
        volume_type1 = self.fake_volume_type()
        cinder_util = cinder_helper.CinderHelper()
        cinder_util.cinder.volume_types.list.return_value = [volume_type1]
        volume_type_name = cinder_util.get_volume_type_by_backendname(
            'nobackend')

        self.assertEqual([], volume_type_name)

    @staticmethod
    def fake_volume(**kwargs):
        volume = mock.MagicMock()
        volume.id = kwargs.get('id', '45a37aeb-95ab-4ddb-a305-7d9f62c2f5ba')
        volume.name = kwargs.get('name', 'fakename')
        volume.size = kwargs.get('size', '1')
        volume.status = kwargs.get('status', 'available')
        volume.snapshot_id = kwargs.get('snapshot_id', None)
        volume.availability_zone = kwargs.get('availability_zone', 'nova')
        volume.volume_type = kwargs.get('volume_type', 'fake_type')
        return volume

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_migrate_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        setattr(volume, 'migration_status', 'success')
        cinder_util.cinder.volumes.get.return_value = volume

        volume_type = self.fake_volume_type()
        cinder_util.cinder.volume_types.list.return_value = [volume_type]

        result = cinder_util.migrate(volume, 'host@backend#pool')
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_migrate_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.cinder.volumes.get.return_value = volume

        volume_type = self.fake_volume_type()
        volume_type.name = 'notbackend'
        cinder_util.cinder.volume_types.list.return_value = [volume_type]

        self.assertRaisesRegex(
            exception.Invalid,
            "Volume type must be same for migrating",
            cinder_util.migrate, volume, 'host@backend#pool')

        volume = self.fake_volume()
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        setattr(volume, 'migration_status', 'error')
        cinder_util.cinder.volumes.get.return_value = volume

        volume_type = self.fake_volume_type()
        cinder_util.cinder.volume_types.list.return_value = [volume_type]

        result = cinder_util.migrate(volume, 'host@backend#pool')
        self.assertFalse(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_retype_success(self, mock_cinder):
        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        setattr(volume, 'migration_status', 'success')
        cinder_util.cinder.volumes.get.return_value = volume

        result = cinder_util.retype(volume, 'notfake_type')
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_retype_fail(self, mock_cinder):
        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        setattr(volume, 'migration_status', 'success')
        cinder_util.cinder.volumes.get.return_value = volume

        self.assertRaisesRegex(
            exception.Invalid,
            "Volume type must be different for retyping",
            cinder_util.retype, volume, 'fake_type')

        volume = self.fake_volume()
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        setattr(volume, 'migration_status', 'error')
        cinder_util.cinder.volumes.get.return_value = volume

        result = cinder_util.retype(volume, 'notfake_type')
        self.assertFalse(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_create_volume_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util.cinder.volumes.create.return_value = volume
        new_vloume = cinder_util.create_volume(
            cinder_util.cinder, volume, 'fake_type')
        self.assertEqual(new_vloume, volume)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_create_volume_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        setattr(volume, 'status', 'fake_status')
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util.cinder.volumes.create.return_value = volume

        self.assertRaisesRegex(
            Exception,
            "Failed to create volume",
            cinder_util.create_volume, cinder_util.cinder, volume,
            'fake_type', retry=2, retry_interval=1)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_delete_volume_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util.cinder.volumes.create.return_value = volume
        cinder_util.check_volume_deleted = mock.MagicMock(return_value=True)
        result = cinder_util.delete_volume(volume)
        self.assertIsNone(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_delete_volume_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        setattr(volume, 'status', 'fake_status')
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util.cinder.volumes.create.return_value = volume
        cinder_util.check_volume_deleted = mock.MagicMock(return_value=False)

        self.assertRaisesRegex(
            Exception,
            "Failed to delete volume",
            cinder_util.delete_volume,
            volume)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_can_get_volume_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.get_volume = mock.MagicMock(return_value=volume)
        result = cinder_util._can_get_volume(volume.id)
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_can_get_volume_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.get_volume = mock.MagicMock()
        cinder_util.get_volume.side_effect =\
            cinder_exception.NotFound(HTTPStatus.NOT_FOUND)
        result = cinder_util._can_get_volume(volume.id)
        self.assertFalse(result)

        cinder_util.get_volume = mock.MagicMock(return_value=None)
        self.assertRaises(
            Exception,
            cinder_util._can_get_volume,
            volume.id)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_has_snapshot_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        volume.snapshot_id = utils.generate_uuid()
        cinder_util.get_volume = mock.MagicMock(return_value=volume)
        result = cinder_util._has_snapshot(volume)
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_has_snapshot_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        volume.snapshot_id = None
        cinder_util.get_volume = mock.MagicMock(return_value=volume)
        result = cinder_util._has_snapshot(volume)
        self.assertFalse(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_get_volume_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.cinder.volumes.get.return_value = volume
        result = cinder_util.get_volume(volume)
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_get_volume_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        side_effect = cinder_exception.NotFound(HTTPStatus.NOT_FOUND)
        cinder_util.cinder.volumes.get.side_effect = side_effect
        cinder_util.cinder.volumes.find.return_value = False
        result = cinder_util.get_volume(volume)
        self.assertFalse(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_check_volume_deleted_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util._can_get_volume = mock.MagicMock(return_value=None)
        result = cinder_util.check_volume_deleted(
            volume, retry=2, retry_interval=1)
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_check_volume_deleted_fail(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util._can_get_volume = mock.MagicMock(return_value=volume)
        result = cinder_util.check_volume_deleted(
            volume, retry=2, retry_interval=1)
        self.assertFalse(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_check_migrated_success(self, mock_cinder):

        cinder_util = cinder_helper.CinderHelper()

        volume = self.fake_volume()
        setattr(volume, 'migration_status', 'success')
        setattr(volume, 'os-vol-host-attr:host', 'host@backend#pool')
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util.check_volume_deleted = mock.MagicMock(return_value=True)
        result = cinder_util.check_migrated(volume)
        self.assertTrue(result)

    @mock.patch.object(time, 'sleep', mock.Mock())
    def test_check_migrated_fail(self, mock_cinder):

        def side_effect(volume):
            if isinstance(volume, str):
                volume = self.fake_volume()
                setattr(volume, 'migration_status', 'error')
            elif volume.id is None:
                setattr(volume, 'migration_status', 'fake_status')
                setattr(volume, 'id', utils.generate_uuid())
            return volume

        cinder_util = cinder_helper.CinderHelper()

        # verify that the method check_migrated will return False when the
        # status of migration_status is error.
        volume = self.fake_volume()
        setattr(volume, 'migration_status', 'error')
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        cinder_util.cinder.volumes.get.return_value = volume
        result = cinder_util.check_migrated(volume, retry_interval=1)
        self.assertFalse(result)

        # verify that the method check_migrated will return False when the
        # status of migration_status is in other cases.
        volume = self.fake_volume()
        setattr(volume, 'migration_status', 'success')
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        setattr(volume, 'id', None)
        cinder_util.get_volume = mock.MagicMock()
        cinder_util.get_volume.side_effect = side_effect
        result = cinder_util.check_migrated(volume, retry_interval=1)
        self.assertFalse(result)

        # verify that the method check_migrated will return False when the
        # return_value of method check_volume_deleted is False.
        volume = self.fake_volume()
        setattr(volume, 'migration_status', 'success')
        setattr(volume, 'os-vol-host-attr:host', 'source_node')
        cinder_util.cinder.volumes.get.return_value = volume
        cinder_util.check_volume_deleted = mock.MagicMock(return_value=False)
        cinder_util.get_deleting_volume = mock.MagicMock(return_value=volume)
        result = cinder_util.check_migrated(volume, retry_interval=1)
        self.assertFalse(result)



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.applier.actions import base as abase
from watcher.applier.loading import default
from watcher.tests import base


class TestDefaultActionLoader(base.TestCase):
    def setUp(self):
        super(TestDefaultActionLoader, self).setUp()
        self.loader = default.DefaultActionLoader()

    def test_endpoints(self):
        for endpoint in self.loader.list_available():
            loaded = self.loader.load(endpoint)
            self.assertIsNotNone(loaded)
            self.assertIsInstance(loaded, abase.BaseAction)



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from unittest import mock

from watcher.applier.actions import base as abase
from watcher.applier.actions import factory
from watcher.applier.workflow_engine import default as tflow
from watcher.common import exception
from watcher.common import utils
from watcher import notifications
from watcher import objects
from watcher.tests.db import base
from watcher.tests.objects import utils as obj_utils


class ExpectedException(Exception):
    pass


class FakeAction(abase.BaseAction, metaclass=abc.ABCMeta):
    def schema(self):
        pass

    def post_condition(self):
        pass

    def pre_condition(self):
        pass

    def revert(self):
        pass

    def execute(self):
        return False

    def get_description(self):
        return "fake action, just for test"


class TestDefaultWorkFlowEngine(base.DbTestCase):
    def setUp(self):
        super(TestDefaultWorkFlowEngine, self).setUp()
        self.engine = tflow.DefaultWorkFlowEngine(
            config=mock.Mock(),
            context=self.context,
            applier_manager=mock.MagicMock())
        self.engine.config.max_workers = 2

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch('taskflow.engines.load')
    @mock.patch('taskflow.patterns.graph_flow.Flow.link')
    def test_execute(self, graph_flow, engines, m_actionplan, m_strategy):
        actions = mock.MagicMock()
        try:
            self.engine.execute(actions)
            self.assertTrue(engines.called)
        except Exception as exc:
            self.fail(exc)

    def create_action(self, action_type, parameters, parents=None, uuid=None,
                      state=None):
        action = {
            'uuid': uuid or utils.generate_uuid(),
            'action_plan_id': 0,
            'action_type': action_type,
            'input_parameters': parameters,
            'state': objects.action.State.PENDING,
            'parents': parents or [],

        }
        new_action = objects.Action(self.context, **action)
        with mock.patch.object(notifications.action, 'send_create'):
            new_action.create()
        return new_action

    def check_action_state(self, action, expected_state):
        to_check = objects.Action.get_by_uuid(self.context, action.uuid)
        self.assertEqual(expected_state, to_check.state)

    def check_actions_state(self, actions, expected_state):
        for a in actions:
            self.check_action_state(a, expected_state)

    @mock.patch('taskflow.engines.load')
    @mock.patch('taskflow.patterns.graph_flow.Flow.link')
    def test_execute_with_no_actions(self, graph_flow, engines):
        actions = []
        try:
            self.engine.execute(actions)
            self.assertFalse(graph_flow.called)
            self.assertTrue(engines.called)
        except Exception as exc:
            self.fail(exc)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    def test_execute_with_one_action(self, mock_send_update,
                                     mock_execution_notification,
                                     m_get_actionplan, m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = [self.create_action("nop", {'message': 'test'})]
        try:
            self.engine.execute(actions)
            self.check_actions_state(actions, objects.action.State.SUCCEEDED)

        except Exception as exc:
            self.fail(exc)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    def test_execute_nop_sleep(self, mock_send_update,
                               mock_execution_notification,
                               m_get_actionplan, m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = []
        first_nop = self.create_action("nop", {'message': 'test'})
        second_nop = self.create_action("nop", {'message': 'second test'})
        sleep = self.create_action("sleep", {'duration': 0.0},
                                   parents=[first_nop.uuid, second_nop.uuid])
        actions.extend([first_nop, second_nop, sleep])

        try:
            self.engine.execute(actions)
            self.check_actions_state(actions, objects.action.State.SUCCEEDED)

        except Exception as exc:
            self.fail(exc)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    def test_execute_with_parents(self, mock_send_update,
                                  mock_execution_notification,
                                  m_get_actionplan, m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = []
        first_nop = self.create_action(
            "nop", {'message': 'test'},
            uuid='bc7eee5c-4fbe-4def-9744-b539be55aa19')
        second_nop = self.create_action(
            "nop", {'message': 'second test'},
            uuid='0565bd5c-aa00-46e5-8d81-2cb5cc1ffa23')
        first_sleep = self.create_action(
            "sleep", {'duration': 0.0}, parents=[first_nop.uuid,
                                                 second_nop.uuid],
            uuid='be436531-0da3-4dad-a9c0-ea1d2aff6496')
        second_sleep = self.create_action(
            "sleep", {'duration': 0.0}, parents=[first_sleep.uuid],
            uuid='9eb51e14-936d-4d12-a500-6ba0f5e0bb1c')
        actions.extend([first_nop, second_nop, first_sleep, second_sleep])

        expected_nodes = [
            {'uuid': 'bc7eee5c-4fbe-4def-9744-b539be55aa19',
             'input_parameters': {u'message': u'test'},
             'action_plan_id': 0, 'state': u'PENDING', 'parents': [],
             'action_type': u'nop', 'id': 1},
            {'uuid': '0565bd5c-aa00-46e5-8d81-2cb5cc1ffa23',
             'input_parameters': {u'message': u'second test'},
             'action_plan_id': 0, 'state': u'PENDING', 'parents': [],
             'action_type': u'nop', 'id': 2},
            {'uuid': 'be436531-0da3-4dad-a9c0-ea1d2aff6496',
             'input_parameters': {u'duration': 0.0},
             'action_plan_id': 0, 'state': u'PENDING',
             'parents': [u'bc7eee5c-4fbe-4def-9744-b539be55aa19',
                         u'0565bd5c-aa00-46e5-8d81-2cb5cc1ffa23'],
             'action_type': u'sleep', 'id': 3},
            {'uuid': '9eb51e14-936d-4d12-a500-6ba0f5e0bb1c',
             'input_parameters': {u'duration': 0.0},
             'action_plan_id': 0, 'state': u'PENDING',
             'parents': [u'be436531-0da3-4dad-a9c0-ea1d2aff6496'],
             'action_type': u'sleep', 'id': 4}]

        expected_edges = [
            ('action_type:nop uuid:0565bd5c-aa00-46e5-8d81-2cb5cc1ffa23',
             'action_type:sleep uuid:be436531-0da3-4dad-a9c0-ea1d2aff6496'),
            ('action_type:nop uuid:bc7eee5c-4fbe-4def-9744-b539be55aa19',
             'action_type:sleep uuid:be436531-0da3-4dad-a9c0-ea1d2aff6496'),
            ('action_type:sleep uuid:be436531-0da3-4dad-a9c0-ea1d2aff6496',
             'action_type:sleep uuid:9eb51e14-936d-4d12-a500-6ba0f5e0bb1c')]

        try:
            flow = self.engine.execute(actions)
            actual_nodes = sorted([x[0]._db_action.as_dict()
                                   for x in flow.iter_nodes()],
                                  key=lambda x: x['id'])
            for expected, actual in zip(expected_nodes, actual_nodes):
                for key in expected.keys():
                    self.assertIn(expected[key], actual.values())
            actual_edges = [(u.name, v.name)
                            for (u, v, _) in flow.iter_links()]

            for edge in expected_edges:
                self.assertIn(edge, actual_edges)

            self.check_actions_state(actions, objects.action.State.SUCCEEDED)

        except Exception as exc:
            self.fail(exc)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    def test_execute_with_two_actions(self, m_send_update, m_execution,
                                      m_get_actionplan, m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = []
        second = self.create_action("sleep", {'duration': 0.0})
        first = self.create_action("nop", {'message': 'test'})

        actions.append(first)
        actions.append(second)

        try:
            self.engine.execute(actions)
            self.check_actions_state(actions, objects.action.State.SUCCEEDED)

        except Exception as exc:
            self.fail(exc)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    def test_execute_with_three_actions(self, m_send_update, m_execution,
                                        m_get_actionplan, m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = []
        third = self.create_action("nop", {'message': 'next'})
        second = self.create_action("sleep", {'duration': 0.0})
        first = self.create_action("nop", {'message': 'hello'})

        self.check_action_state(first, objects.action.State.PENDING)
        self.check_action_state(second, objects.action.State.PENDING)
        self.check_action_state(third, objects.action.State.PENDING)

        actions.append(first)
        actions.append(second)
        actions.append(third)

        try:
            self.engine.execute(actions)
            self.check_actions_state(actions, objects.action.State.SUCCEEDED)

        except Exception as exc:
            self.fail(exc)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    def test_execute_with_exception(self, m_send_update, m_execution,
                                    m_get_actionplan, m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = []

        third = self.create_action("no_exist", {'message': 'next'})
        second = self.create_action("sleep", {'duration': 0.0})
        first = self.create_action("nop", {'message': 'hello'})

        self.check_action_state(first, objects.action.State.PENDING)
        self.check_action_state(second, objects.action.State.PENDING)
        self.check_action_state(third, objects.action.State.PENDING)

        actions.append(first)
        actions.append(second)
        actions.append(third)

        self.engine.execute(actions)

        self.check_action_state(first, objects.action.State.SUCCEEDED)
        self.check_action_state(second, objects.action.State.SUCCEEDED)
        self.check_action_state(third, objects.action.State.FAILED)

    @mock.patch.object(objects.Strategy, "get_by_id")
    @mock.patch.object(objects.ActionPlan, "get_by_id")
    @mock.patch.object(notifications.action, 'send_execution_notification')
    @mock.patch.object(notifications.action, 'send_update')
    @mock.patch.object(factory.ActionFactory, "make_action")
    def test_execute_with_action_failed(self, m_make_action, m_send_update,
                                        m_send_execution, m_get_actionplan,
                                        m_get_strategy):
        m_get_actionplan.return_value = obj_utils.get_test_action_plan(
            self.context, id=0)
        m_get_strategy.return_value = obj_utils.get_test_strategy(
            self.context, id=1)
        actions = [self.create_action("fake_action", {})]
        m_make_action.return_value = FakeAction(mock.Mock())

        self.engine.execute(actions)
        self.check_action_state(actions[0], objects.action.State.FAILED)

    @mock.patch.object(objects.ActionPlan, "get_by_uuid")
    def test_execute_with_action_plan_cancel(self, m_get_actionplan):
        obj_utils.create_test_goal(self.context)
        strategy = obj_utils.create_test_strategy(self.context)
        audit = obj_utils.create_test_audit(
            self.context, strategy_id=strategy.id)
        action_plan = obj_utils.create_test_action_plan(
            self.context, audit_id=audit.id,
            strategy_id=strategy.id,
            state=objects.action_plan.State.CANCELLING)
        action1 = obj_utils.create_test_action(
            self.context, action_plan_id=action_plan.id,
            action_type='nop', state=objects.action.State.SUCCEEDED,
            input_parameters={'message': 'hello World'})
        action2 = obj_utils.create_test_action(
            self.context, action_plan_id=action_plan.id,
            action_type='nop', state=objects.action.State.ONGOING,
            uuid='9eb51e14-936d-4d12-a500-6ba0f5e0bb1c',
            input_parameters={'message': 'hello World'})
        action3 = obj_utils.create_test_action(
            self.context, action_plan_id=action_plan.id,
            action_type='nop', state=objects.action.State.PENDING,
            uuid='bc7eee5c-4fbe-4def-9744-b539be55aa19',
            input_parameters={'message': 'hello World'})
        m_get_actionplan.return_value = action_plan
        actions = []
        actions.append(action1)
        actions.append(action2)
        actions.append(action3)
        self.assertRaises(exception.ActionPlanCancelled,
                          self.engine.execute, actions)
        try:
            self.check_action_state(action1, objects.action.State.SUCCEEDED)
            self.check_action_state(action2, objects.action.State.CANCELLED)
            self.check_action_state(action3, objects.action.State.CANCELLED)

        except Exception as exc:
            self.fail(exc)

    def test_decider(self):
        # execution_rule is ALWAYS
        self.engine.execution_rule = 'ALWAYS'
        history = {'action1': True}
        self.assertTrue(self.engine.decider(history))

        history = {'action1': False}
        self.assertTrue(self.engine.decider(history))

        # execution_rule is ANY
        self.engine.execution_rule = 'ANY'
        history = {'action1': True}
        self.assertFalse(self.engine.decider(history))

        history = {'action1': False}
        self.assertTrue(self.engine.decider(history))



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

from watcher.common import exception
from watcher.db.sqlalchemy import api as db_api
from watcher import objects
from watcher.tests.db import base
from watcher.tests.db import utils


class TestStrategyObject(base.DbTestCase):

    goal_id = 2

    scenarios = [
        ('non_eager', dict(
            eager=False, fake_strategy=utils.get_test_strategy(
                goal_id=goal_id))),
        ('eager_with_non_eager_load', dict(
            eager=True, fake_strategy=utils.get_test_strategy(
                goal_id=goal_id))),
        ('eager_with_eager_load', dict(
            eager=True, fake_strategy=utils.get_test_strategy(
                goal_id=goal_id, goal=utils.get_test_goal(id=goal_id)))),
    ]

    def setUp(self):
        super(TestStrategyObject, self).setUp()
        self.fake_goal = utils.create_test_goal(id=self.goal_id)

    def eager_load_strategy_assert(self, strategy):
        if self.eager:
            self.assertIsNotNone(strategy.goal)
            fields_to_check = set(
                super(objects.Goal, objects.Goal).fields
            ).symmetric_difference(objects.Goal.fields)
            db_data = {
                k: v for k, v in self.fake_goal.as_dict().items()
                if k in fields_to_check}
            object_data = {
                k: v for k, v in strategy.goal.as_dict().items()
                if k in fields_to_check}
            self.assertEqual(db_data, object_data)

    @mock.patch.object(db_api.Connection, 'get_strategy_by_id')
    def test_get_by_id(self, mock_get_strategy):
        strategy_id = self.fake_strategy['id']
        mock_get_strategy.return_value = self.fake_strategy
        strategy = objects.Strategy.get(
            self.context, strategy_id, eager=self.eager)
        mock_get_strategy.assert_called_once_with(
            self.context, strategy_id, eager=self.eager)
        self.assertEqual(self.context, strategy._context)
        self.eager_load_strategy_assert(strategy)

    @mock.patch.object(db_api.Connection, 'get_strategy_by_uuid')
    def test_get_by_uuid(self, mock_get_strategy):
        uuid = self.fake_strategy['uuid']
        mock_get_strategy.return_value = self.fake_strategy
        strategy = objects.Strategy.get(self.context, uuid, eager=self.eager)
        mock_get_strategy.assert_called_once_with(
            self.context, uuid, eager=self.eager)
        self.assertEqual(self.context, strategy._context)
        self.eager_load_strategy_assert(strategy)

    def test_get_bad_uuid(self):
        self.assertRaises(exception.InvalidIdentity,
                          objects.Strategy.get, self.context, 'not-a-uuid')

    @mock.patch.object(db_api.Connection, 'get_strategy_list')
    def test_list(self, mock_get_list):
        mock_get_list.return_value = [self.fake_strategy]
        strategies = objects.Strategy.list(self.context, eager=self.eager)
        self.assertEqual(1, mock_get_list.call_count, 1)
        self.assertEqual(1, len(strategies))
        self.assertIsInstance(strategies[0], objects.Strategy)
        self.assertEqual(self.context, strategies[0]._context)
        for strategy in strategies:
            self.eager_load_strategy_assert(strategy)

    @mock.patch.object(db_api.Connection, 'update_strategy')
    @mock.patch.object(db_api.Connection, 'get_strategy_by_id')
    def test_save(self, mock_get_strategy, mock_update_strategy):
        _id = self.fake_strategy['id']
        mock_get_strategy.return_value = self.fake_strategy
        strategy = objects.Strategy.get_by_id(
            self.context, _id, eager=self.eager)
        strategy.name = 'UPDATED NAME'
        strategy.save()

        mock_get_strategy.assert_called_once_with(
            self.context, _id, eager=self.eager)
        mock_update_strategy.assert_called_once_with(
            _id, {'name': 'UPDATED NAME'})
        self.assertEqual(self.context, strategy._context)
        self.eager_load_strategy_assert(strategy)

    @mock.patch.object(db_api.Connection, 'get_strategy_by_id')
    def test_refresh(self, mock_get_strategy):
        _id = self.fake_strategy['id']
        returns = [dict(self.fake_strategy, name="first name"),
                   dict(self.fake_strategy, name="second name")]
        mock_get_strategy.side_effect = returns
        expected = [mock.call(self.context, _id, eager=self.eager),
                    mock.call(self.context, _id, eager=self.eager)]
        strategy = objects.Strategy.get(self.context, _id, eager=self.eager)
        self.assertEqual("first name", strategy.name)
        strategy.refresh(eager=self.eager)
        self.assertEqual("second name", strategy.name)
        self.assertEqual(expected, mock_get_strategy.call_args_list)
        self.assertEqual(self.context, strategy._context)
        self.eager_load_strategy_assert(strategy)


class TestCreateDeleteStrategyObject(base.DbTestCase):

    def setUp(self):
        super(TestCreateDeleteStrategyObject, self).setUp()
        self.fake_goal = utils.create_test_goal()
        self.fake_strategy = utils.get_test_strategy(goal_id=self.fake_goal.id)

    @mock.patch.object(db_api.Connection, 'create_strategy')
    def test_create(self, mock_create_strategy):
        mock_create_strategy.return_value = self.fake_strategy
        strategy = objects.Strategy(self.context, **self.fake_strategy)
        strategy.create()
        mock_create_strategy.assert_called_once_with(self.fake_strategy)
        self.assertEqual(self.context, strategy._context)

    @mock.patch.object(db_api.Connection, 'soft_delete_strategy')
    @mock.patch.object(db_api.Connection, 'get_strategy_by_id')
    def test_soft_delete(self, mock_get_strategy, mock_soft_delete):
        _id = self.fake_strategy['id']
        mock_get_strategy.return_value = self.fake_strategy
        strategy = objects.Strategy.get_by_id(self.context, _id)
        strategy.soft_delete()
        mock_get_strategy.assert_called_once_with(
            self.context, _id, eager=False)
        mock_soft_delete.assert_called_once_with(_id)
        self.assertEqual(self.context, strategy._context)

    @mock.patch.object(db_api.Connection, 'destroy_strategy')
    @mock.patch.object(db_api.Connection, 'get_strategy_by_id')
    def test_destroy(self, mock_get_strategy, mock_destroy_strategy):
        _id = self.fake_strategy['id']
        mock_get_strategy.return_value = self.fake_strategy
        strategy = objects.Strategy.get_by_id(self.context, _id)
        strategy.destroy()
        mock_get_strategy.assert_called_once_with(
            self.context, _id, eager=False)
        mock_destroy_strategy.assert_called_once_with(_id)
        self.assertEqual(self.context, strategy._context)



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 Servionica
#
# Authors: Alexander Chadin <a.chadin@servionica.ru>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from watcher.common import context as wcontext
from watcher.common import exception
from watcher.notifications import action_plan as ap_notifications
from watcher.notifications import base as notificationbase
from watcher.notifications import exception as exception_notifications
from watcher import objects
from watcher.objects import base
from watcher.objects import fields as wfields

CONF = cfg.CONF


@base.WatcherObjectRegistry.register_notification
class ActionPayload(notificationbase.NotificationPayloadBase):
    SCHEMA = {
        'uuid': ('action', 'uuid'),

        'action_type': ('action', 'action_type'),
        'input_parameters': ('action', 'input_parameters'),
        'state': ('action', 'state'),
        'parents': ('action', 'parents'),

        'created_at': ('action', 'created_at'),
        'updated_at': ('action', 'updated_at'),
        'deleted_at': ('action', 'deleted_at'),
    }

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'uuid': wfields.UUIDField(),
        'action_type': wfields.StringField(nullable=False),
        'input_parameters': wfields.DictField(nullable=False, default={}),
        'state': wfields.StringField(nullable=False),
        'parents': wfields.ListOfUUIDsField(nullable=False, default=[]),
        'action_plan_uuid': wfields.UUIDField(),
        'action_plan': wfields.ObjectField('TerseActionPlanPayload'),

        'created_at': wfields.DateTimeField(nullable=True),
        'updated_at': wfields.DateTimeField(nullable=True),
        'deleted_at': wfields.DateTimeField(nullable=True),
    }

    def __init__(self, action, **kwargs):
        super(ActionPayload, self).__init__(**kwargs)
        self.populate_schema(action=action)


@base.WatcherObjectRegistry.register_notification
class ActionStateUpdatePayload(notificationbase.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'old_state': wfields.StringField(nullable=True),
        'state': wfields.StringField(nullable=True),
    }


@base.WatcherObjectRegistry.register_notification
class ActionCreatePayload(ActionPayload):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {}

    def __init__(self, action, action_plan):
        super(ActionCreatePayload, self).__init__(
            action=action,
            action_plan=action_plan)


@base.WatcherObjectRegistry.register_notification
class ActionUpdatePayload(ActionPayload):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'state_update': wfields.ObjectField('ActionStateUpdatePayload'),
    }

    def __init__(self, action, state_update, action_plan):
        super(ActionUpdatePayload, self).__init__(
            action=action,
            state_update=state_update,
            action_plan=action_plan)


@base.WatcherObjectRegistry.register_notification
class ActionExecutionPayload(ActionPayload):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'fault': wfields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, action, action_plan, **kwargs):
        super(ActionExecutionPayload, self).__init__(
            action=action,
            action_plan=action_plan,
            **kwargs)


@base.WatcherObjectRegistry.register_notification
class ActionCancelPayload(ActionPayload):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'fault': wfields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, action, action_plan, **kwargs):
        super(ActionCancelPayload, self).__init__(
            action=action,
            action_plan=action_plan,
            **kwargs)


@base.WatcherObjectRegistry.register_notification
class ActionDeletePayload(ActionPayload):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {}

    def __init__(self, action, action_plan):
        super(ActionDeletePayload, self).__init__(
            action=action,
            action_plan=action_plan)


@notificationbase.notification_sample('action-execution-error.json')
@notificationbase.notification_sample('action-execution-end.json')
@notificationbase.notification_sample('action-execution-start.json')
@base.WatcherObjectRegistry.register_notification
class ActionExecutionNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionExecutionPayload')
    }


@notificationbase.notification_sample('action-create.json')
@base.WatcherObjectRegistry.register_notification
class ActionCreateNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionCreatePayload')
    }


@notificationbase.notification_sample('action-update.json')
@base.WatcherObjectRegistry.register_notification
class ActionUpdateNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionUpdatePayload')
    }


@notificationbase.notification_sample('action-delete.json')
@base.WatcherObjectRegistry.register_notification
class ActionDeleteNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionDeletePayload')
    }


@notificationbase.notification_sample('action-cancel-error.json')
@notificationbase.notification_sample('action-cancel-end.json')
@notificationbase.notification_sample('action-cancel-start.json')
@base.WatcherObjectRegistry.register_notification
class ActionCancelNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionCancelPayload')
    }


def _get_action_plan_payload(action):
    action_plan = None
    strategy_uuid = None
    audit = None
    try:
        action_plan = action.action_plan
        audit = objects.Audit.get(wcontext.make_context(show_deleted=True),
                                  action_plan.audit_id)
        if audit.strategy_id:
            strategy_uuid = objects.Strategy.get(
                wcontext.make_context(show_deleted=True),
                audit.strategy_id).uuid
    except NotImplementedError:
        raise exception.EagerlyLoadedActionRequired(action=action.uuid)

    action_plan_payload = ap_notifications.TerseActionPlanPayload(
        action_plan=action_plan,
        audit_uuid=audit.uuid, strategy_uuid=strategy_uuid)

    return action_plan_payload


def send_create(context, action, service='infra-optim', host=None):
    """Emit an action.create notification."""
    action_plan_payload = _get_action_plan_payload(action)

    versioned_payload = ActionCreatePayload(
        action=action,
        action_plan=action_plan_payload,
    )

    notification = ActionCreateNotification(
        priority=wfields.NotificationPriority.INFO,
        event_type=notificationbase.EventType(
            object='action',
            action=wfields.NotificationAction.CREATE),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_update(context, action, service='infra-optim',
                host=None, old_state=None):
    """Emit an action.update notification."""
    action_plan_payload = _get_action_plan_payload(action)

    state_update = ActionStateUpdatePayload(
        old_state=old_state,
        state=action.state if old_state else None)

    versioned_payload = ActionUpdatePayload(
        action=action,
        state_update=state_update,
        action_plan=action_plan_payload,
    )

    notification = ActionUpdateNotification(
        priority=wfields.NotificationPriority.INFO,
        event_type=notificationbase.EventType(
            object='action',
            action=wfields.NotificationAction.UPDATE),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_delete(context, action, service='infra-optim', host=None):
    """Emit an action.delete notification."""
    action_plan_payload = _get_action_plan_payload(action)

    versioned_payload = ActionDeletePayload(
        action=action,
        action_plan=action_plan_payload,
    )

    notification = ActionDeleteNotification(
        priority=wfields.NotificationPriority.INFO,
        event_type=notificationbase.EventType(
            object='action',
            action=wfields.NotificationAction.DELETE),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_execution_notification(context, action, notification_action, phase,
                                priority=wfields.NotificationPriority.INFO,
                                service='infra-optim', host=None):
    """Emit an action execution notification."""
    action_plan_payload = _get_action_plan_payload(action)

    fault = None
    if phase == wfields.NotificationPhase.ERROR:
        fault = exception_notifications.ExceptionPayload.from_exception()

    versioned_payload = ActionExecutionPayload(
        action=action,
        action_plan=action_plan_payload,
        fault=fault,
    )

    notification = ActionExecutionNotification(
        priority=priority,
        event_type=notificationbase.EventType(
            object='action',
            action=notification_action,
            phase=phase),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_cancel_notification(context, action, notification_action, phase,
                             priority=wfields.NotificationPriority.INFO,
                             service='infra-optim', host=None):
    """Emit an action cancel notification."""
    action_plan_payload = _get_action_plan_payload(action)

    fault = None
    if phase == wfields.NotificationPhase.ERROR:
        fault = exception_notifications.ExceptionPayload.from_exception()

    versioned_payload = ActionCancelPayload(
        action=action,
        action_plan=action_plan_payload,
        fault=fault,
    )

    notification = ActionCancelNotification(
        priority=priority,
        event_type=notificationbase.EventType(
            object='action',
            action=notification_action,
            phase=phase),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An efficacy specification is a contract that is associated to each :ref:`Goal
<goal_definition>` that defines the various :ref:`efficacy indicators
<efficacy_indicator_definition>` a strategy achieving the associated goal
should provide within its :ref:`solution <solution_definition>`. Indeed, each
solution proposed by a strategy will be validated against this contract before
calculating its :ref:`global efficacy <efficacy_definition>`.
"""

import abc
import jsonschema
from oslo_serialization import jsonutils


class EfficacySpecification(object, metaclass=abc.ABCMeta):

    def __init__(self):
        self._indicators_specs = self.get_indicators_specifications()

    @property
    def indicators_specs(self):
        return self._indicators_specs

    @abc.abstractmethod
    def get_indicators_specifications(self):
        """List the specifications of the indicator for this efficacy spec

        :return: Tuple of indicator specifications
        :rtype: Tuple of :py:class:`~.IndicatorSpecification` instances
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_global_efficacy_indicator(self, indicators_map):
        """Compute the global efficacy for the goal it achieves

        :param indicators_map: dict-like object containing the
                               efficacy indicators related to this spec
        :type indicators_map: :py:class:`~.IndicatorsMap` instance
        :raises: NotImplementedError
        :returns: :py:class:`~.Indicator` instance list, each instance specify
                  global efficacy for different openstack resource.
        """
        raise NotImplementedError()

    @property
    def schema(self):
        """Combined schema from the schema of the indicators"""
        schema = {
            "type": "object",
            "properties": {},
            "required": []
        }
        for indicator in self.indicators_specs:
            schema["properties"][indicator.name] = indicator.schema
            if indicator.required:
                schema["required"].append(indicator.name)
        return schema

    def validate_efficacy_indicators(self, indicators_map):
        if indicators_map:
            jsonschema.validate(indicators_map, self.schema)
        else:
            True

    def get_indicators_specs_dicts(self):
        return [indicator.to_dict()
                for indicator in self.indicators_specs]

    def serialize_indicators_specs(self):
        return jsonutils.dumps(self.get_indicators_specs_dicts())



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
from unittest import mock

from oslo_config import cfg
from oslo_service import wsgi
from pecan.testing import load_test_app

from watcher.api import config as api_config
from watcher.cmd import api
from watcher.common import service
from watcher.tests import base


class TestApi(base.BaseTestCase):

    def setUp(self):
        super(TestApi, self).setUp()

        self.conf = cfg.CONF
        self._parse_cli_opts = self.conf._parse_cli_opts

        def _fake_parse(self, args=[]):
            return cfg.ConfigOpts._parse_cli_opts(self, [])

        _fake_parse_method = types.MethodType(_fake_parse, self.conf)
        self.conf._parse_cli_opts = _fake_parse_method

    def tearDown(self):
        super(TestApi, self).tearDown()
        self.conf._parse_cli_opts = self._parse_cli_opts

    @mock.patch.object(wsgi, "Server", mock.Mock())
    @mock.patch("watcher.api.app.pecan.make_app")
    @mock.patch.object(service, "launch")
    def test_run_api_app(self, m_launcher, m_make_app):
        m_make_app.return_value = load_test_app(config=api_config.PECAN_CONFIG)
        api.main()
        self.assertEqual(1, m_launcher.call_count)

    @mock.patch.object(wsgi, "Server", mock.Mock())
    @mock.patch("watcher.api.app.pecan.make_app")
    @mock.patch.object(service, "launch")
    def test_run_api_app_serve_specific_address(self, m_launcher, m_make_app):
        cfg.CONF.set_default("host", "localhost", group="api")
        m_make_app.return_value = load_test_app(config=api_config.PECAN_CONFIG)
        api.main()
        self.assertEqual(1, m_launcher.call_count)



# -*- encoding: utf-8 -*-
# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.common import exception
from watcher.common import utils
from watcher.db import api as db_api
from watcher.objects import base
from watcher.objects import fields as wfields


@base.WatcherObjectRegistry.register
class EfficacyIndicator(base.WatcherPersistentObject, base.WatcherObject,
                        base.WatcherObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = db_api.get_instance()

    fields = {
        'id': wfields.IntegerField(),
        'uuid': wfields.UUIDField(),
        'action_plan_id': wfields.IntegerField(),
        'name': wfields.StringField(),
        'description': wfields.StringField(nullable=True),
        'unit': wfields.StringField(nullable=True),
        'value': wfields.NumericField(),
    }

    @base.remotable_classmethod
    def get(cls, context, efficacy_indicator_id):
        """Find an efficacy indicator object given its ID or UUID

        :param efficacy_indicator_id: the ID or UUID of an efficacy indicator.
        :returns: a :class:`EfficacyIndicator` object.
        """
        if utils.is_int_like(efficacy_indicator_id):
            return cls.get_by_id(context, efficacy_indicator_id)
        elif utils.is_uuid_like(efficacy_indicator_id):
            return cls.get_by_uuid(context, efficacy_indicator_id)
        else:
            raise exception.InvalidIdentity(identity=efficacy_indicator_id)

    @base.remotable_classmethod
    def get_by_id(cls, context, efficacy_indicator_id):
        """Find an efficacy indicator given its integer ID

        :param efficacy_indicator_id: the id of an efficacy indicator.
        :returns: a :class:`EfficacyIndicator` object.
        """
        db_efficacy_indicator = cls.dbapi.get_efficacy_indicator_by_id(
            context, efficacy_indicator_id)
        efficacy_indicator = EfficacyIndicator._from_db_object(
            cls(context), db_efficacy_indicator)
        return efficacy_indicator

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        """Find an efficacy indicator given its UUID

        :param uuid: the uuid of an efficacy indicator.
        :param context: Security context
        :returns: a :class:`EfficacyIndicator` object.
        """
        db_efficacy_indicator = cls.dbapi.get_efficacy_indicator_by_uuid(
            context, uuid)
        efficacy_indicator = EfficacyIndicator._from_db_object(
            cls(context), db_efficacy_indicator)
        return efficacy_indicator

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None, filters=None,
             sort_key=None, sort_dir=None):
        """Return a list of EfficacyIndicator objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param filters: Filters to apply. Defaults to None.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :returns: a list of :class:`EfficacyIndicator` object.

        """
        db_efficacy_indicators = cls.dbapi.get_efficacy_indicator_list(
            context,
            limit=limit,
            marker=marker,
            filters=filters,
            sort_key=sort_key,
            sort_dir=sort_dir)

        return [cls._from_db_object(cls(context), obj)
                for obj in db_efficacy_indicators]

    @base.remotable
    def create(self, context=None):
        """Create a EfficacyIndicator record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: EfficacyIndicator(context)

        """
        values = self.obj_get_changes()
        db_efficacy_indicator = self.dbapi.create_efficacy_indicator(values)
        self._from_db_object(self, db_efficacy_indicator)

    def destroy(self, context=None):
        """Delete the EfficacyIndicator from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: EfficacyIndicator(context)
        """
        self.dbapi.destroy_efficacy_indicator(self.uuid)
        self.obj_reset_changes()

    @base.remotable
    def save(self, context=None):
        """Save updates to this EfficacyIndicator.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: EfficacyIndicator(context)
        """
        updates = self.obj_get_changes()
        self.dbapi.update_efficacy_indicator(self.uuid, updates)

        self.obj_reset_changes()

    @base.remotable
    def refresh(self, context=None):
        """Loads updates for this EfficacyIndicator.

        Loads an efficacy indicator with the same uuid from the database and
        checks for updated attributes. Updates are applied to the loaded
        efficacy indicator column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: EfficacyIndicator(context)
        """
        current = self.__class__.get_by_uuid(self._context, uuid=self.uuid)
        self.obj_refresh(current)

    @base.remotable
    def soft_delete(self, context=None):
        """Soft Delete the efficacy indicator from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Audit(context)
        """
        self.dbapi.soft_delete_efficacy_indicator(self.uuid)



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel
#
# Authors: Tomasz Kaczynski <tomasz.kaczynski@intel.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import random

from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import units

from watcher._i18n import _
from watcher.decision_engine.scoring import scoring_factory
from watcher.decision_engine.strategy.strategies import base

LOG = log.getLogger(__name__)


class DummyWithScorer(base.DummyBaseStrategy):
    """A dummy strategy using dummy scoring engines.

    This is a dummy strategy demonstrating how to work with scoring
    engines. One scoring engine is predicting the workload type of a machine
    based on the telemetry data, the other one is simply calculating the
    average value for given elements in a list. Results are then passed to the
    NOP action.

    The strategy is presenting the whole workflow:
    - Get a reference to a scoring engine
    - Prepare input data (features) for score calculation
    - Perform score calculation
    - Use scorer's metadata for results interpretation
    """

    DEFAULT_NAME = "dummy_with_scorer"
    DEFAULT_DESCRIPTION = "Dummy Strategy with Scorer"

    NOP = "nop"
    SLEEP = "sleep"

    def __init__(self, config, osc=None):
        """Constructor: the signature should be identical within the subclasses

        :param config: Configuration related to this plugin
        :type config: :py:class:`~.Struct`
        :param osc: An OpenStackClients instance
        :type osc: :py:class:`~.OpenStackClients` instance
        """

        super(DummyWithScorer, self).__init__(config, osc)

        # Setup Scoring Engines
        self._workload_scorer = (scoring_factory
                                 .get_scoring_engine('dummy_scorer'))
        self._avg_scorer = (scoring_factory
                            .get_scoring_engine('dummy_avg_scorer'))

        # Get metainfo from Workload Scorer for result interpretation
        metainfo = jsonutils.loads(self._workload_scorer.get_metainfo())
        self._workloads = {index: workload
                           for index, workload in enumerate(
                               metainfo['workloads'])}

    def pre_execute(self):
        self._pre_execute()

    def do_execute(self, audit=None):
        # Simple "hello world" from strategy
        param1 = self.input_parameters.param1
        param2 = self.input_parameters.param2
        LOG.debug('DummyWithScorer params: param1=%(p1)f, param2=%(p2)s',
                  {'p1': param1, 'p2': param2})
        parameters = {'message': 'Hello from Dummy Strategy with Scorer!'}
        self.solution.add_action(action_type=self.NOP,
                                 input_parameters=parameters)

        # Demonstrate workload scorer
        features = self._generate_random_telemetry()
        result_str = self._workload_scorer.calculate_score(features)
        LOG.debug('Workload Scorer result: %s', result_str)

        # Parse the result using workloads from scorer's metainfo
        result = self._workloads[jsonutils.loads(result_str)[0]]
        LOG.debug('Detected Workload: %s', result)
        parameters = {'message': 'Detected Workload: %s' % result}
        self.solution.add_action(action_type=self.NOP,
                                 input_parameters=parameters)

        # Demonstrate AVG scorer
        features = jsonutils.dumps(random.sample(range(1000), 20))
        result_str = self._avg_scorer.calculate_score(features)
        LOG.debug('AVG Scorer result: %s', result_str)
        result = jsonutils.loads(result_str)[0]
        LOG.debug('AVG Scorer result (parsed): %d', result)
        parameters = {'message': 'AVG Scorer result: %s' % result}
        self.solution.add_action(action_type=self.NOP,
                                 input_parameters=parameters)

        # Sleep action
        self.solution.add_action(action_type=self.SLEEP,
                                 input_parameters={'duration': 5.0})

    def post_execute(self):
        pass

    @classmethod
    def get_name(cls):
        return 'dummy_with_scorer'

    @classmethod
    def get_display_name(cls):
        return _('Dummy Strategy using sample Scoring Engines')

    @classmethod
    def get_translatable_display_name(cls):
        return 'Dummy Strategy using sample Scoring Engines'

    @classmethod
    def get_schema(cls):
        # Mandatory default setting for each element
        return {
            'properties': {
                'param1': {
                    'description': 'number parameter example',
                    'type': 'number',
                    'default': 3.2,
                    'minimum': 1.0,
                    'maximum': 10.2,
                },
                'param2': {
                    'description': 'string parameter example',
                    'type': "string",
                    'default': "hello"
                },
            },
        }

    def _generate_random_telemetry(self):
        processor_time = random.randint(0, 100)
        mem_total_bytes = 4*units.Gi
        mem_avail_bytes = random.randint(1*units.Gi, 4*units.Gi)
        mem_page_reads = random.randint(0, 2000)
        mem_page_writes = random.randint(0, 2000)
        disk_read_bytes = random.randint(0*units.Mi, 200*units.Mi)
        disk_write_bytes = random.randint(0*units.Mi, 200*units.Mi)
        net_bytes_received = random.randint(0*units.Mi, 20*units.Mi)
        net_bytes_sent = random.randint(0*units.Mi, 10*units.Mi)

        return jsonutils.dumps([
            processor_time, mem_total_bytes, mem_avail_bytes,
            mem_page_reads, mem_page_writes, disk_read_bytes,
            disk_write_bytes, net_bytes_received, net_bytes_sent])



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_log import log

from watcher.applier.actions import base

LOG = log.getLogger(__name__)


class Nop(base.BaseAction):
    """logs a message

    The action schema is::

        schema = Schema({
         'message': str,
        })

    The `message` is the actual message that will be logged.
    """

    MESSAGE = 'message'

    @property
    def schema(self):
        return {
            'type': 'object',
            'properties': {
                'message': {
                    'type': ['string', 'null']
                }
            },
            'required': ['message'],
            'additionalProperties': False,
        }

    @property
    def message(self):
        return self.input_parameters.get(self.MESSAGE)

    def execute(self):
        LOG.debug("Executing action NOP message: %s ", self.message)
        return True

    def revert(self):
        LOG.debug("Revert action NOP")
        return True

    def pre_condition(self):
        pass

    def post_condition(self):
        pass

    def get_description(self):
        """Description of the action"""
        return "Logging a NOP message"

    def abort(self):
        LOG.debug("Abort action NOP")
        return True



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
# Copyright (c) 2016 Intel Corp
#
# Authors: Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from watcher.conf import api
from watcher.conf import applier
from watcher.conf import ceilometer_client
from watcher.conf import cinder_client
from watcher.conf import clients_auth
from watcher.conf import collector
from watcher.conf import datasources
from watcher.conf import db
from watcher.conf import decision_engine
from watcher.conf import exception
from watcher.conf import glance_client
from watcher.conf import gnocchi_client
from watcher.conf import grafana_client
from watcher.conf import grafana_translators
from watcher.conf import ironic_client
from watcher.conf import keystone_client
from watcher.conf import monasca_client
from watcher.conf import neutron_client
from watcher.conf import nova_client
from watcher.conf import paths
from watcher.conf import placement_client
from watcher.conf import planner
from watcher.conf import service

CONF = cfg.CONF

service.register_opts(CONF)
api.register_opts(CONF)
paths.register_opts(CONF)
exception.register_opts(CONF)
datasources.register_opts(CONF)
db.register_opts(CONF)
planner.register_opts(CONF)
applier.register_opts(CONF)
decision_engine.register_opts(CONF)
monasca_client.register_opts(CONF)
nova_client.register_opts(CONF)
glance_client.register_opts(CONF)
gnocchi_client.register_opts(CONF)
keystone_client.register_opts(CONF)
grafana_client.register_opts(CONF)
grafana_translators.register_opts(CONF)
cinder_client.register_opts(CONF)
ceilometer_client.register_opts(CONF)
neutron_client.register_opts(CONF)
clients_auth.register_opts(CONF)
ironic_client.register_opts(CONF)
collector.register_opts(CONF)
placement_client.register_opts(CONF)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


placement_group = cfg.OptGroup(
    'placement_client',
    title='Placement Service Options',
    help="Configuration options for connecting to the placement API service")

placement_opts = [
    cfg.StrOpt('api_version',
               default='1.29',
               help='microversion of placement API when using '
                    'placement service.'),
    cfg.StrOpt('interface',
               default='public',
               choices=['internal', 'public', 'admin'],
               help='Type of endpoint when using placement service.'),
    cfg.StrOpt('region_name',
               help='Region in Identity service catalog to use for '
                    'communication with the OpenStack service.')]


def register_opts(conf):
    conf.register_group(placement_group)
    conf.register_opts(placement_opts, group=placement_group)


def list_opts():
    return [(placement_group, placement_opts)]



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities and helper functions."""

import datetime
import random
import re
import string

from croniter import croniter

from jsonschema import validators
from oslo_config import cfg
from oslo_log import log
from oslo_utils import strutils
from oslo_utils import uuidutils

from watcher.common import exception

CONF = cfg.CONF

LOG = log.getLogger(__name__)


class Struct(dict):
    """Specialized dict where you access an item like an attribute

    >>> struct = Struct()
    >>> struct['a'] = 1
    >>> struct.b = 2
    >>> assert struct.a == 1
    >>> assert struct['b'] == 2
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        try:
            self[name] = value
        except KeyError:
            raise AttributeError(name)


generate_uuid = uuidutils.generate_uuid
is_uuid_like = uuidutils.is_uuid_like
is_int_like = strutils.is_int_like


def is_cron_like(value):
    """Return True is submitted value is like cron syntax"""
    try:
        croniter(value, datetime.datetime.now())
    except Exception as e:
        raise exception.CronFormatIsInvalid(message=str(e))
    return True


def safe_rstrip(value, chars=None):
    """Removes trailing characters from a string if that does not make it empty

    :param value: A string value that will be stripped.
    :param chars: Characters to remove.
    :return: Stripped value.

    """
    if not isinstance(value, str):
        LOG.warning(
            "Failed to remove trailing character. Returning original object."
            "Supplied object is not a string: %s,", value)
        return value

    return value.rstrip(chars) or value


def is_hostname_safe(hostname):
    """Determine if the supplied hostname is RFC compliant.

    Check that the supplied hostname conforms to:
        * http://en.wikipedia.org/wiki/Hostname
        * http://tools.ietf.org/html/rfc952
        * http://tools.ietf.org/html/rfc1123

    :param hostname: The hostname to be validated.
    :returns: True if valid. False if not.

    """
    m = r'^[a-z0-9]([a-z0-9\-]{0,61}[a-z0-9])?$'
    return (isinstance(hostname, str) and
            (re.match(m, hostname) is not None))


def get_cls_import_path(cls):
    """Return the import path of a given class"""
    module = cls.__module__
    if module is None or module == str.__module__:
        return cls.__name__
    return module + '.' + cls.__name__


# Default value feedback extension as jsonschema doesn't support it
def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for prop, subschema in properties.items():
            if "default" in subschema and instance is not None:
                instance.setdefault(prop, subschema["default"])

            for error in validate_properties(
                validator, properties, instance, schema
            ):
                yield error

    return validators.extend(validator_class,
                             {"properties": set_defaults})


# Parameter strict check extension as jsonschema doesn't support it
def extend_with_strict_schema(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def strict_schema(validator, properties, instance, schema):
        if instance is None:
            return

        for para in instance.keys():
            if para not in properties.keys():
                raise exception.AuditParameterNotAllowed(parameter=para)

            for error in validate_properties(
                validator, properties, instance, schema
            ):
                yield error

    return validators.extend(validator_class, {"properties": strict_schema})


StrictDefaultValidatingDraft4Validator = extend_with_default(
    extend_with_strict_schema(validators.Draft4Validator))

Draft4Validator = validators.Draft4Validator


def random_string(n):
    return ''.join([random.choice(
        string.ascii_letters + string.digits) for i in range(n)])



# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher.applier.action_plan import default
from watcher.applier import default as ap_applier
from watcher.common import exception
from watcher import notifications
from watcher import objects
from watcher.objects import action_plan as ap_objects
from watcher.tests.db import base
from watcher.tests.objects import utils as obj_utils


class TestDefaultActionPlanHandler(base.DbTestCase):

    class FakeApplierException(Exception):
        pass

    def setUp(self):
        super(TestDefaultActionPlanHandler, self).setUp()

        p_action_plan_notifications = mock.patch.object(
            notifications, 'action_plan', autospec=True)
        self.m_action_plan_notifications = p_action_plan_notifications.start()
        self.addCleanup(p_action_plan_notifications.stop)

        obj_utils.create_test_goal(self.context)
        self.strategy = obj_utils.create_test_strategy(self.context)
        self.audit = obj_utils.create_test_audit(
            self.context, strategy_id=self.strategy.id)
        self.action_plan = obj_utils.create_test_action_plan(
            self.context, audit_id=self.audit.id,
            strategy_id=self.strategy.id)
        self.action = obj_utils.create_test_action(
            self.context, action_plan_id=self.action_plan.id,
            action_type='nop',
            input_parameters={'message': 'hello World'})

    @mock.patch.object(objects.ActionPlan, "get_by_uuid")
    def test_launch_action_plan(self, m_get_action_plan):
        m_get_action_plan.return_value = self.action_plan
        command = default.DefaultActionPlanHandler(
            self.context, mock.MagicMock(), self.action_plan.uuid)
        command.execute()

        expected_calls = [
            mock.call(self.context, self.action_plan,
                      action=objects.fields.NotificationAction.EXECUTION,
                      phase=objects.fields.NotificationPhase.START),
            mock.call(self.context, self.action_plan,
                      action=objects.fields.NotificationAction.EXECUTION,
                      phase=objects.fields.NotificationPhase.END)]

        self.assertEqual(ap_objects.State.SUCCEEDED, self.action_plan.state)

        self.assertEqual(
            expected_calls,
            self.m_action_plan_notifications
                .send_action_notification
                .call_args_list)

    @mock.patch.object(ap_applier.DefaultApplier, "execute")
    @mock.patch.object(objects.ActionPlan, "get_by_uuid")
    def test_launch_action_plan_with_error(self, m_get_action_plan, m_execute):
        m_get_action_plan.return_value = self.action_plan
        m_execute.side_effect = self.FakeApplierException
        command = default.DefaultActionPlanHandler(
            self.context, mock.MagicMock(), self.action_plan.uuid)
        command.execute()

        expected_calls = [
            mock.call(self.context, self.action_plan,
                      action=objects.fields.NotificationAction.EXECUTION,
                      phase=objects.fields.NotificationPhase.START),
            mock.call(self.context, self.action_plan,
                      action=objects.fields.NotificationAction.EXECUTION,
                      priority=objects.fields.NotificationPriority.ERROR,
                      phase=objects.fields.NotificationPhase.ERROR)]

        self.assertEqual(ap_objects.State.FAILED, self.action_plan.state)

        self.assertEqual(
            expected_calls,
            self.m_action_plan_notifications
                .send_action_notification
                .call_args_list)

    @mock.patch.object(objects.ActionPlan, "get_by_uuid")
    def test_cancel_action_plan(self, m_get_action_plan):
        m_get_action_plan.return_value = self.action_plan
        self.action_plan.state = ap_objects.State.CANCELLED
        self.action_plan.save()
        command = default.DefaultActionPlanHandler(
            self.context, mock.MagicMock(), self.action_plan.uuid)
        command.execute()
        action = self.action.get_by_uuid(self.context, self.action.uuid)
        self.assertEqual(ap_objects.State.CANCELLED, self.action_plan.state)
        self.assertEqual(objects.action.State.CANCELLED, action.state)

    @mock.patch.object(ap_applier.DefaultApplier, "execute")
    @mock.patch.object(objects.ActionPlan, "get_by_uuid")
    def test_cancel_action_plan_with_exception(self, m_get_action_plan,
                                               m_execute):
        m_get_action_plan.return_value = self.action_plan
        m_execute.side_effect = exception.ActionPlanCancelled(
            self.action_plan.uuid)
        command = default.DefaultActionPlanHandler(
            self.context, mock.MagicMock(), self.action_plan.uuid)
        command.execute()
        self.assertEqual(ap_objects.State.CANCELLED, self.action_plan.state)
